from typing import Dict, Optional
from unittest import mock

import pytest

from opentrons_shared_data.pipette import name_for_model
from opentrons_shared_data.pipette.types import PipetteModel

from opentrons import types
from opentrons.hardware_control import ot3api
from opentrons.hardware_control.types import Axis


@pytest.mark.parametrize(
    "pipette_model", ["p1000_single_v3.3", "p1000_single_v3.3", "p50_multi_v3.3"]
)
async def test_transforms_roundtrip(pipette_model: PipetteModel) -> None:
    attached: Dict[types.Mount, Dict[str, Optional[str]]] = {
        types.Mount.LEFT: {
            "model": pipette_model,
            "id": pipette_model + "_idididid_left",
            "name": name_for_model(pipette_model),
        },
        types.Mount.RIGHT: {
            "model": pipette_model,
            "id": pipette_model + "_idididid_right",
            "name": name_for_model(pipette_model),
        },
    }
    sim = await ot3api.OT3API.build_hardware_simulator(attached_instruments=attached)
    target = types.Point(20, 30, 40)
    await sim.move_to(types.Mount.RIGHT, target)
    assert target == await sim.gantry_position(types.Mount.RIGHT)


@pytest.mark.parametrize(
    "pipette_model", ["p1000_single_v3.3", "p50_single_v3.3", "p1000_multi_v3.3"]
)
async def test_transform_values(
    pipette_model: PipetteModel, enable_ot3_hardware_controller: None
) -> None:
    attached: Dict[types.Mount, Dict[str, Optional[str]]] = {
        types.Mount.LEFT: {
            "model": pipette_model,
            "id": pipette_model + "_idididid_left",
            "name": name_for_model(pipette_model),
        },
        types.Mount.RIGHT: {
            "model": pipette_model,
            "id": pipette_model + "_idididid_right",
            "name": name_for_model(pipette_model),
        },
    }
    sim = await ot3api.OT3API.build_hardware_simulator(attached_instruments=attached)
    target = types.Point(20, 30, 40)
    # mypy has some support for mock.patch.object but not for magicmock
    with mock.patch.object(  # type: ignore[call-overload]
        sim._backend,
        "move",
        mock.MagicMock(side_effect=sim._backend.move),
        spec=sim._backend.move,
    ) as mock_move:
        await sim.move_to(types.Mount.RIGHT, target)
        right_pipette = sim.hardware_instruments[types.Mount.RIGHT]
        assert right_pipette
        right_offset = right_pipette.critical_point()
        point = [
            (target.x - right_offset[0] - sim.config.right_mount_offset[0]) * -1
            + sim.config.carriage_offset[0],
            (target.y - right_offset[1] - sim.config.right_mount_offset[1]) * -1
            + sim.config.carriage_offset[1],
            (target.z - right_offset[2] - sim.config.right_mount_offset[2]) * -1
            + sim.config.carriage_offset[2],
        ]
        assert mock_move.call_args[0][1][Axis.X] == point[0]
        assert mock_move.call_args[0][1][Axis.Y] == point[1]
        assert mock_move.call_args[0][1][Axis.Z_R] == point[2]

    # mypy has some support for mock.patch.object but not for magicmock
    with mock.patch.object(  # type: ignore[call-overload]
        sim._backend,
        "move",
        mock.MagicMock(side_effect=sim._backend.move),
        spec=sim._backend.move,
    ) as mock_move:
        await sim.move_to(types.Mount.LEFT, target)
        left_pipette = sim.hardware_instruments[types.Mount.LEFT]
        assert left_pipette
        left_offset = left_pipette.critical_point()
        point = [
            (target.x - left_offset[0] - sim.config.left_mount_offset[0]) * -1
            + sim.config.carriage_offset[0],
            (target.y - left_offset[1] - sim.config.left_mount_offset[1]) * -1
            + sim.config.carriage_offset[1],
            (target.z - left_offset[2] - sim.config.left_mount_offset[2]) * -1
            + sim.config.carriage_offset[2],
        ]
        assert mock_move.call_args[0][1][Axis.X] == point[0]
        assert mock_move.call_args[0][1][Axis.Y] == point[1]
        assert mock_move.call_args[0][1][Axis.Z_L] == point[2]



"""Tests for TaskQueue."""

import asyncio

from decoy import Decoy

from opentrons.protocol_runner.task_queue import TaskQueue


async def test_set_run_func(decoy: Decoy) -> None:
    """It should be able to add a task for the "run" phase."""
    run_func = decoy.mock(name="run_func", is_async=True)
    cleanup_func = decoy.mock(name="cleanup_func", is_async=True)

    subject = TaskQueue()  # cleanup_func=cleanup_func)
    subject.set_cleanup_func(func=cleanup_func)
    subject.set_run_func(func=run_func)
    subject.start()
    await subject.join()

    decoy.verify(
        await run_func(),
        await cleanup_func(None),
    )


async def test_passes_args(decoy: Decoy) -> None:
    """It should pass kwargs to the run phase function."""
    run_func = decoy.mock(name="run_func", is_async=True)
    cleanup_func = decoy.mock(name="cleanup_func", is_async=True)

    subject = TaskQueue()  # cleanup_func=cleanup_func)
    subject.set_cleanup_func(func=cleanup_func)
    subject.set_run_func(func=run_func, hello="world")
    subject.start()
    await subject.join()

    decoy.verify(await run_func(hello="world"))


async def test_cleanup_gets_run_error(decoy: Decoy) -> None:
    """It should verify "cleanup" func gets error raised in "run" func."""
    run_func = decoy.mock(name="run_func", is_async=True)
    cleanup_func = decoy.mock(name="cleanup_func", is_async=True)
    error = RuntimeError("Oh no!")

    decoy.when(await run_func()).then_raise(error)

    subject = TaskQueue()  # cleanup_func=cleanup_func)
    subject.set_cleanup_func(func=cleanup_func)
    subject.set_run_func(func=run_func)
    subject.start()
    await subject.join()

    decoy.verify(await cleanup_func(error))


async def test_join_waits_for_start(decoy: Decoy) -> None:
    """It should wait until the queue is started when join is called."""
    cleanup_func = decoy.mock(name="cleanup_func", is_async=True)
    subject = TaskQueue()  # cleanup_func=cleanup_func)
    subject.set_cleanup_func(func=cleanup_func)
    join_task = asyncio.create_task(subject.join())

    await asyncio.sleep(0)
    assert join_task.done() is False

    subject.start()
    await join_task


async def test_start_runs_stuff_once(decoy: Decoy) -> None:
    """Calling `start` should no-op if already started."""
    run_func = decoy.mock(name="run_func", is_async=True)
    cleanup_func = decoy.mock(name="cleanup_func", is_async=True)

    subject = TaskQueue()  # leanup_func=cleanup_func)
    subject.set_cleanup_func(func=cleanup_func)
    subject.set_run_func(func=run_func)
    subject.start()
    subject.start()
    await subject.join()

    decoy.verify(await run_func(), times=1)
    decoy.verify(await cleanup_func(None), times=1)



"""Command models to engage a user to empty a Flex Stacker."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Type

from ...errors import (
    ErrorOccurrence,
)
from ...errors.exceptions import FlexStackerLabwarePoolNotYetDefinedError
from ...state import update_types
from ...types import (
    OFF_DECK_LOCATION,
    InStackerHopperLocation,
    LabwareLocation,
    LabwareLocationSequence,
    NotOnDeckLocationSequenceComponent,
    OnLabwareLocation,
    StackerFillEmptyStrategy,
    StackerStoredLabwareGroup,
)
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from .common import (
    adapter_location_sequences_with_default,
    labware_locations_for_group,
    lid_location_sequences_with_default,
    primary_location_sequences,
)

if TYPE_CHECKING:
    from ...execution import EquipmentHandler, RunControlHandler
    from ...state.state import StateView

EmptyCommandType = Literal["flexStacker/empty"]


class EmptyParams(BaseModel):
    """The parameters defining how a stacker should be emptied."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")

    strategy: StackerFillEmptyStrategy = Field(
        ...,
        description=(
            "How to empty the stacker. "
            "If manualWithPause, pause the protocol until the client sends an interaction, and mark "
            "the labware pool as empty thereafter. If logical, do not pause but immediately apply the "
            "specified count."
        ),
    )

    message: str | SkipJsonSchema[None] = Field(
        None,
        description="The message to display on connected clients during a manualWithPause strategy empty.",
    )

    count: Optional[Annotated[int, Field(ge=0)]] = Field(
        None,
        description=(
            "The new count of labware in the pool. If None, default to an empty pool. If this number is "
            "larger than the amount of labware currently in the pool, default to the smaller amount. "
            "Do not use the value in the parameters as an outside observer; instead, use the count value "
            "from the results."
        ),
    )


class EmptyResult(BaseModel):
    """Result data from a stacker empty command."""

    count: int = Field(
        ..., description="The new amount of labware stored in the stacker labware pool."
    )
    primaryLabwareURI: str = Field(
        ...,
        description="The labware definition URI of the primary labware.",
    )
    adapterLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the adapter labware.",
    )
    lidLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the lid labware.",
    )
    storedLabware: list[StackerStoredLabwareGroup] | SkipJsonSchema[None] = Field(
        ..., description="The primary labware loaded into the stacker labware pool."
    )
    removedLabware: list[StackerStoredLabwareGroup] | SkipJsonSchema[None] = Field(
        ...,
        description="The labware objects that have just been removed from the stacker labware pool.",
    )
    originalPrimaryLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each newly-removed primary labware, in the same order as removedLabware.",
    )
    originalAdapterLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each newly-removed adapter labware, in the same order as removedLabware. None if the pool does not specify an adapter.",
    )
    originalLidLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each newly-removed lid labware, in the same order as removedLabware. None if the  pool does not specify a lid.",
    )
    newPrimaryLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each newly-removed primary labware, in the same order as removedLabware.",
    )
    newAdapterLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each newly-removed adapter labware, in the same order as removedLabware. None if the pool does not specify an adapter.",
    )
    newLidLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each newly-removed lid labware, in the same order as removedLabware. None if the pool does not specify a lid labware.",
    )


class EmptyImpl(AbstractCommandImpl[EmptyParams, SuccessData[EmptyResult]]):
    """Implementation of a stacker empty command."""

    def __init__(
        self,
        state_view: StateView,
        run_control: RunControlHandler,
        equipment: EquipmentHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._run_control = run_control
        self._equipment = equipment

    async def execute(  # noqa: C901
        self, params: EmptyParams
    ) -> SuccessData[EmptyResult]:
        """Execute the stacker empty command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )

        if stacker_state.pool_primary_definition is None:
            location = self._state_view.modules.get_location(params.moduleId)
            raise FlexStackerLabwarePoolNotYetDefinedError(
                message=f"The Flex Stacker in {location} has not been configured yet and cannot be emptied."
            )

        count = params.count if params.count is not None else 0

        new_count = min(len(stacker_state.contained_labware_bottom_first), count)

        new_stored_labware = stacker_state.contained_labware_bottom_first[:new_count]
        removed_labware = stacker_state.contained_labware_bottom_first[new_count:]
        new_locations_by_id: dict[str, LabwareLocation] = {}
        new_offset_ids_by_id: dict[str, str | None] = {}

        def _add_to_dicts(labware_group: StackerStoredLabwareGroup) -> None:
            if labware_group.adapterLabwareId:
                new_locations_by_id[labware_group.primaryLabwareId] = OnLabwareLocation(
                    labwareId=labware_group.adapterLabwareId
                )
                new_locations_by_id[labware_group.adapterLabwareId] = OFF_DECK_LOCATION
                new_offset_ids_by_id[labware_group.primaryLabwareId] = None
                new_offset_ids_by_id[labware_group.adapterLabwareId] = None
            else:
                new_locations_by_id[labware_group.primaryLabwareId] = OFF_DECK_LOCATION
                new_offset_ids_by_id[labware_group.primaryLabwareId] = None
            if labware_group.lidLabwareId:
                new_locations_by_id[labware_group.lidLabwareId] = OnLabwareLocation(
                    labwareId=labware_group.primaryLabwareId
                )
                new_offset_ids_by_id[labware_group.lidLabwareId] = None

        for group in removed_labware:
            _add_to_dicts(group)

        state_update = (
            update_types.StateUpdate()
            .update_flex_stacker_contained_labware(
                module_id=params.moduleId,
                contained_labware_bottom_first=new_stored_labware,
            )
            .set_batch_labware_location(
                new_locations_by_id=new_locations_by_id,
                new_offset_ids_by_id=new_offset_ids_by_id,
            )
        )

        stacker_hw = self._equipment.get_module_hardware_api(
            module_id=stacker_state.module_id
        )
        if stacker_hw:
            stacker_hw.set_stacker_identify(True)

        if params.strategy == StackerFillEmptyStrategy.MANUAL_WITH_PAUSE:
            await self._run_control.wait_for_resume()

        if stacker_hw:
            stacker_hw.set_stacker_identify(False)

        if stacker_state.pool_primary_definition is None:
            raise FlexStackerLabwarePoolNotYetDefinedError(
                "The Primary Labware must be defined in the stacker pool."
            )

        original_locations = [
            labware_locations_for_group(
                group, [InStackerHopperLocation(moduleId=params.moduleId)]
            )
            for group in removed_labware
        ]
        new_locations = [
            labware_locations_for_group(
                group,
                [
                    NotOnDeckLocationSequenceComponent(
                        logicalLocationName=OFF_DECK_LOCATION
                    )
                ],
            )
            for group in removed_labware
        ]

        return SuccessData(
            public=EmptyResult.model_construct(
                count=new_count,
                primaryLabwareURI=self._state_view.labware.get_uri_from_definition(
                    stacker_state.pool_primary_definition
                ),
                adapterLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_adapter_definition
                ),
                lidLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_lid_definition
                ),
                storedLabware=new_stored_labware,
                removedLabware=removed_labware,
                originalPrimaryLabwareLocationSequences=primary_location_sequences(
                    original_locations
                ),
                originalAdapterLabwareLocationSequences=adapter_location_sequences_with_default(
                    original_locations, stacker_state.pool_adapter_definition
                ),
                originalLidLabwareLocationSequences=lid_location_sequences_with_default(
                    original_locations, stacker_state.pool_lid_definition
                ),
                newPrimaryLabwareLocationSequences=primary_location_sequences(
                    new_locations
                ),
                newAdapterLabwareLocationSequences=adapter_location_sequences_with_default(
                    new_locations, stacker_state.pool_adapter_definition
                ),
                newLidLabwareLocationSequences=lid_location_sequences_with_default(
                    new_locations, stacker_state.pool_lid_definition
                ),
            ),
            state_update=state_update,
        )


class Empty(BaseCommand[EmptyParams, EmptyResult, ErrorOccurrence]):
    """A command to empty the Flex Stacker of labware."""

    commandType: EmptyCommandType = "flexStacker/empty"
    params: EmptyParams
    result: Optional[EmptyResult] = None

    _ImplementationCls: Type[EmptyImpl] = EmptyImpl


class EmptyCreate(BaseCommandCreate[EmptyParams]):
    """A request to execute a Flex Stacker empty command."""

    commandType: EmptyCommandType = "flexStacker/empty"
    params: EmptyParams

    _CommandCls: Type[Empty] = Empty



"""Test move relative commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError

from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.move_relative import (
    MoveRelativeImplementation,
    MoveRelativeParams,
    MoveRelativeResult,
)
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.execution import MovementHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.types import DeckPoint, MovementAxis
from opentrons.types import Point


@pytest.fixture
def subject(
    movement: MovementHandler, model_utils: ModelUtils
) -> MoveRelativeImplementation:
    """Build a MoveRelativeImplementation with injected dependencies."""
    return MoveRelativeImplementation(movement=movement, model_utils=model_utils)


async def test_move_relative_implementation(
    decoy: Decoy, movement: MovementHandler, subject: MoveRelativeImplementation
) -> None:
    """A MoveRelative command should have an execution implementation."""
    data = MoveRelativeParams(
        pipetteId="pipette-id",
        axis=MovementAxis.X,
        distance=42.0,
    )

    decoy.when(
        await movement.move_relative(
            pipette_id="pipette-id",
            axis=MovementAxis.X,
            distance=42.0,
        )
    ).then_return(Point(x=1, y=2, z=3))

    result = await subject.execute(data)

    assert result == SuccessData(
        public=MoveRelativeResult(position=DeckPoint(x=1, y=2, z=3)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=update_types.NO_CHANGE,
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            )
        ),
    )


async def test_move_relative_stalls(
    decoy: Decoy,
    movement: MovementHandler,
    model_utils: ModelUtils,
    subject: MoveRelativeImplementation,
) -> None:
    """A MoveRelative command should handle stalls."""
    data = MoveRelativeParams(pipetteId="pipette-id", axis=MovementAxis.Y, distance=40)

    decoy.when(
        await movement.move_relative(
            pipette_id="pipette-id", axis=MovementAxis.Y, distance=40
        )
    ).then_raise(StallOrCollisionDetectedError())

    timestamp = datetime.now()
    test_id = "test-id"

    decoy.when(model_utils.get_timestamp()).then_return(timestamp)
    decoy.when(model_utils.generate_id()).then_return(test_id)

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=test_id, createdAt=timestamp, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )



"""Test cli execution."""

import json
import tempfile
import textwrap
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional

import pytest
from click.testing import CliRunner

from opentrons.cli.analyze import AnalysisResult, analyze


def _list_fixtures(version: int) -> Iterator[Path]:
    return Path(__file__).parent.glob(
        f"../../../../shared-data/protocol/fixtures/{version}/*.json"
    )


@dataclass
class _AnalysisCLIResult:
    exit_code: int
    json_output: Optional[Dict[str, Any]]
    stdout_stderr: str


def _get_analysis_result(
    protocol_files: List[Path],
    output_type: str,
    check: bool = False,
    rtp_values: Optional[str] = None,
    rtp_files: Optional[str] = None,
) -> _AnalysisCLIResult:
    """Run `protocol_files` as a single protocol through the analysis CLI.

    Returns:
        A tuple (exit_code, analysis_json_dict_or_none).

        Don't forget to check the status code. Errors from within the analysis CLI will otherwise
        not be propagated!
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        analysis_output_file = Path(temp_dir) / "analysis_output.json"
        runner = CliRunner()
        args = [output_type, str(analysis_output_file)]

        if rtp_values is not None:
            args.extend(["--rtp-values", rtp_values])

        if rtp_files is not None:
            args.extend(["--rtp-files", rtp_files])

        args.extend([str(p.resolve()) for p in protocol_files])

        if check:
            args.append("--check")

        result = runner.invoke(analyze, args)
        if analysis_output_file.exists():
            json_output = json.loads(analysis_output_file.read_bytes())
        else:
            json_output = None
        return _AnalysisCLIResult(
            exit_code=result.exit_code,
            json_output=json_output,
            stdout_stderr=result.output,
        )


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
@pytest.mark.parametrize("fixture_path", _list_fixtures(6))
def test_analyze(
    fixture_path: Path,
    output: str,
) -> None:
    """Should return with no errors and a non-empty output."""
    result = _get_analysis_result([fixture_path], output)

    assert result.exit_code == 0

    assert result.json_output is not None
    assert "robotType" in result.json_output
    assert "pipettes" in result.json_output
    assert "commands" in result.json_output
    assert "labware" in result.json_output
    assert "liquids" in result.json_output
    assert "modules" in result.json_output
    assert "result" in result.json_output


_DECK_DEFINITION_TEST_SLOT = 2
_DECK_DEFINITION_TEST_LABWARE = "agilent_1_reservoir_290ml"
_DECK_DEFINITION_TEST_WELL = "A1"


def _get_deck_definition_test_source(api_level: str, robot_type: str) -> str:
    return textwrap.dedent(
        f"""\
        requirements = {{
            "apiLevel": "{api_level}",
            "robotType": "{robot_type}",
        }}
        def run(protocol):
            labware = protocol.load_labware(
                "{_DECK_DEFINITION_TEST_LABWARE}",
                "{_DECK_DEFINITION_TEST_SLOT}",
            )
            test_point = labware["{_DECK_DEFINITION_TEST_WELL}"].top().point
            protocol.comment(str(test_point))
        """
    )


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
@pytest.mark.parametrize(
    ("api_level", "robot_type", "expected_point"),
    [
        # These expected_point values were copied from known-good analysis outputs.
        # The exact values don't matter much for this test, since we're not checking positional
        # accuracy here. They just need to be clearly different between the OT-2 and OT-3.
        ("2.13", "OT-2", "(196.38, 42.785, 44.04)"),
        ("2.15", "OT-2", "(196.38, 42.785, 44.04)"),
        pytest.param(
            "2.15",
            "OT-3",
            "(227.88, 42.785, 44.04)",
            marks=pytest.mark.ot3_only,  # Analyzing an OT-3 protocol requires an OT-3 hardware API.
        ),
    ],
)
def test_analysis_deck_definition(
    api_level: str,
    robot_type: str,
    expected_point: str,
    tmp_path: Path,
    output: str,
) -> None:
    """Test that the analysis uses the appropriate deck definition for the protocol's robot type.

    At the time of writing, the only official, public, documented way to observe the deck definition
    that a protocol uses is for the protocol to load a labware and inspect the deck coordinates of
    its wells.
    """
    protocol_source_file = Path(tmp_path) / "protocol.py"
    protocol_source_file.write_text(
        _get_deck_definition_test_source(
            api_level=api_level,
            robot_type=robot_type,
        ),
        encoding="utf-8",
    )

    result = _get_analysis_result([protocol_source_file], output)

    assert result.exit_code == 0

    assert result.json_output is not None
    [home_command, load_labware_command, comment_command] = result.json_output[
        "commands"
    ]

    assert comment_command["params"]["message"] == expected_point


# TODO(mm, 2023-08-12): We can remove this test when we remove special handling for these
# protocols. https://opentrons.atlassian.net/browse/RSS-306
@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_strict_metatada_requirements_validation(tmp_path: Path, output: str) -> None:
    """It should apply strict validation to the metadata and requirements dicts.

    It should reject protocols with questionable metadata and requirements dicts,
    even though these protocols may be accepted by other parts of the system.
    https://opentrons.atlassian.net/browse/RSS-306
    """
    protocol_source = textwrap.dedent(
        """
        # apiLevel in both metadata and requirements
        metadata = {"apiLevel": "2.15"}
        requirements = {"apiLevel": "2.15"}

        def run(protocol):
            pass
        """
    )

    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(protocol_source, encoding="utf-8")

    result = _get_analysis_result([protocol_source_file], output)

    assert result.exit_code != 0

    expected_message = (
        "You may only put apiLevel in the metadata dict or the requirements dict"
    )
    assert expected_message in result.stdout_stderr


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
@pytest.mark.parametrize("check", [True, False])
@pytest.mark.parametrize(
    ("python_protocol_source", "expected_detail"),
    [
        (
            textwrap.dedent(
                # Raises an exception from outside of Opentrons code,
                # in between two PAPI functions.
                """\
                requirements = {"apiLevel": "2.14"}  # line 1
                                                     # line 2
                def run(protocol):                   # line 3
                    protocol.comment(":^)")          # line 4
                    raise RuntimeError(">:(")        # line 5
                    protocol.comment(":D")           # line 6
                """
            ),
            "RuntimeError [line 5]: >:(",
        ),
        (
            textwrap.dedent(
                # Raises an exception from inside a Protocol Engine command.
                # https://opentrons.atlassian.net/browse/RSS-317
                """\
                requirements = {"apiLevel": "2.14"}      # line 1
                                                         # line 2
                def run(protocol):                       # line 3
                    tip_rack = protocol.load_labware(    # line 4
                        "opentrons_96_tiprack_300ul", 1  # line 5
                    )                                    # line 6
                    pipette = protocol.load_instrument(  # line 7
                        "p300_single", "left"            # line 8
                    )                                    # line 9
                    pipette.pick_up_tip(tip_rack["A1"])  # line 10
                    pipette.pick_up_tip(tip_rack["A2"])  # line 11
                """
            ),
            (
                # TODO(mm, 2023-09-12): This is an overly verbose concatenative Frankenstein
                # message. We should simplify our error propagation to trim out the noise.
                "ProtocolCommandFailedError [line 11]:"
                " Error 4000 GENERAL_ERROR (ProtocolCommandFailedError):"
                " TipAttachedError: Pipette should not have a tip attached, but does."
            ),
        ),
        # TODO: PAPIv<2.15?
    ],
)
def test_python_error_line_numbers(
    tmp_path: Path,
    python_protocol_source: str,
    expected_detail: str,
    output: str,
    check: bool,
) -> None:
    """Test that error messages from Python protocols have line numbers."""
    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(python_protocol_source, encoding="utf-8")

    result = _get_analysis_result([protocol_source_file], output, check)

    if check:
        assert result.exit_code != 0
    else:
        assert result.exit_code == 0
    assert result.json_output is not None
    assert result.json_output["result"] == AnalysisResult.NOT_OK.value
    [error] = result.json_output["errors"]
    assert error["detail"] == expected_detail


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_run_time_parameter_setting(
    tmp_path: Path,
    output: str,
) -> None:
    """Test that a RTP can be set to a non default value for analysis.

    Also verify that analysis result contains all static data about the protocol.
    """
    python_protocol_source = textwrap.dedent(
        """\
            requirements = {"robotType": "OT-2", "apiLevel": "2.18"}

            def add_parameters(parameters):
                parameters.add_bool(
                    display_name="Dry Run",
                    variable_name="dry_run",
                    default=False,
                )
            def run(protocol):
                pass
        """
    )
    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(python_protocol_source, encoding="utf-8")
    result = _get_analysis_result(
        [protocol_source_file], output, rtp_values=json.dumps({"dry_run": True})
    )

    assert result.exit_code == 0

    assert result.json_output is not None
    assert result.json_output["robotType"] == "OT-2 Standard"
    assert result.json_output["result"] == AnalysisResult.OK
    assert result.json_output["pipettes"] == []
    assert result.json_output["commands"]  # There should be a home command
    assert result.json_output["labware"] == []
    assert result.json_output["liquids"] == []
    assert result.json_output["modules"] == []
    assert result.json_output["config"] == {
        "apiVersion": [2, 18],
        "protocolType": "python",
    }
    assert result.json_output["files"] == [{"name": "protocol.py", "role": "main"}]
    assert result.json_output["runTimeParameters"] == [
        {
            "displayName": "Dry Run",
            "variableName": "dry_run",
            "type": "bool",
            "value": True,
            "default": False,
        }
    ]


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_run_time_parameter_error(
    tmp_path: Path,
    output: str,
) -> None:
    """Test that an RTP validation error is shown correctly in analysis result.

    Also verify that analysis result contains all static data about the protocol.
    """
    python_protocol_source = textwrap.dedent(
        # Raises an exception during runner load.
        """\
            requirements = {"robotType": "OT-2", "apiLevel": "2.18"}  # line 1
                                                                      # line 2
            def add_parameters(parameters):                           # line 3
                # No default value specified                          # line 4
                parameters.add_bool(                                  # line 5
                    display_name="Dry Run",
                    variable_name="dry_run",
                )
            def run(protocol):
                pass
        """
    )
    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(python_protocol_source, encoding="utf-8")
    result = _get_analysis_result([protocol_source_file], output)

    assert result.exit_code == 0

    assert result.json_output is not None
    assert result.json_output["robotType"] == "OT-2 Standard"
    assert result.json_output["result"] == AnalysisResult.NOT_OK.value
    assert result.json_output["pipettes"] == []
    assert result.json_output["commands"] == []
    assert result.json_output["labware"] == []
    assert result.json_output["liquids"] == []
    assert result.json_output["modules"] == []
    assert result.json_output["config"] == {
        "apiVersion": [2, 18],
        "protocolType": "python",
    }
    assert result.json_output["files"] == [{"name": "protocol.py", "role": "main"}]
    [error] = result.json_output["errors"]
    assert error["detail"] == (
        "TypeError [line 5]: ParameterContext.add_bool() missing 1"
        " required positional argument: 'default'"
    )


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_rtp_csv_file_setting(
    tmp_path: Path,
    output: str,
) -> None:
    """Test that a CSV file can be set for analysis.

    Also verify that analysis result contains all static data about the protocol.
    """
    python_protocol_source = textwrap.dedent(
        """\
            requirements = {"robotType": "OT-2", "apiLevel": "2.20"}

            def add_parameters(parameters):
                parameters.add_csv_file(
                    display_name="CSV File",
                    variable_name="csv_file",
                )
            def run(protocol):
                protocol.params.csv_file.contents
        """
    )
    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(python_protocol_source, encoding="utf-8")
    csv_source_file = tmp_path / "csv_file.csv"
    csv_source_file.write_text("a,b,c", encoding="utf-8")

    result = _get_analysis_result(
        [protocol_source_file],
        output,
        rtp_files=json.dumps({"csv_file": str(csv_source_file.resolve())}),
    )

    assert result.exit_code == 0

    assert result.json_output is not None
    assert result.json_output["robotType"] == "OT-2 Standard"
    assert result.json_output["result"] == AnalysisResult.OK
    assert result.json_output["pipettes"] == []
    assert result.json_output["commands"]  # There should be a home command
    assert result.json_output["labware"] == []
    assert result.json_output["liquids"] == []
    assert result.json_output["modules"] == []
    assert result.json_output["config"] == {
        "apiVersion": [2, 20],
        "protocolType": "python",
    }
    assert result.json_output["files"] == [{"name": "protocol.py", "role": "main"}]
    assert result.json_output["runTimeParameters"] == [
        {
            "displayName": "CSV File",
            "variableName": "csv_file",
            "type": "csv_file",
            "file": {"id": "", "name": "csv_file.csv"},
        }
    ]


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_file_required_error(
    tmp_path: Path,
    output: str,
) -> None:
    """Test that a FileParameterRequired error gets caught and changes the result to FILE_REQUIRED.

    Also verify that analysis result contains all static data about the protocol.
    """
    python_protocol_source = textwrap.dedent(
        # Raises an exception during runner load.
        """\
            requirements = {"robotType": "OT-2", "apiLevel": "2.20"}

            def add_parameters(parameters):
                parameters.add_csv_file(
                    display_name="CSV File",
                    variable_name="csv_file",
                )
            def run(protocol):
                protocol.params.csv_file.file
        """
    )
    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(python_protocol_source, encoding="utf-8")
    result = _get_analysis_result([protocol_source_file], output)

    assert result.exit_code == 0

    assert result.json_output is not None
    assert result.json_output["robotType"] == "OT-2 Standard"
    assert result.json_output["result"] == AnalysisResult.PARAMETER_VALUE_REQUIRED.value
    assert result.json_output["pipettes"] == []
    assert result.json_output["commands"]  # There should be a home command
    assert result.json_output["labware"] == []
    assert result.json_output["liquids"] == []
    assert result.json_output["modules"] == []
    assert result.json_output["config"] == {
        "apiVersion": [2, 20],
        "protocolType": "python",
    }
    assert result.json_output["files"] == [{"name": "protocol.py", "role": "main"}]
    assert result.json_output["errors"]


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_unexpected_error(
    tmp_path: Path,
    output: str,
) -> None:
    """Test that an unexpected error raised from outside opentrons functions is handled correctly."""
    python_protocol_source = textwrap.dedent(
        # Raises an exception before runner load.
        """\
            requirements = {"robotType": "OT-2", "apiLevel": "2.18"}  # line 1
            x + 1 = 0                                                 # line 2
            def add_parameters(parameters):
                parameters.add_bool()
            def run(protocol):
                pass
        """
    )
    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(python_protocol_source, encoding="utf-8")
    result = _get_analysis_result([protocol_source_file], output)

    assert result.exit_code != 0
    assert result.stdout_stderr == (
        "Error: cannot assign to expression here."
        " Maybe you meant '==' instead of '='? (protocol.py, line 2)\n"
    )


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_unexpected_runner_load_error(
    tmp_path: Path,
    output: str,
) -> None:
    """Test that an error raised during runner load is handled properly.

    Also verify that analysis result contains all static data about the protocol.
    """
    python_protocol_source = textwrap.dedent(
        # Raises an exception during runner load.
        """\
            requirements = {"apiLevel": "2.18"}     # line 1
            call_a_non_existent_func()              # line 2

            def add_parameters(parameters):         # line 4
                parameters.add_bool()
            def run(protocol):
                pass
        """
    )
    protocol_source_file = tmp_path / "protocol.py"
    protocol_source_file.write_text(python_protocol_source, encoding="utf-8")
    result = _get_analysis_result([protocol_source_file], output)

    assert result.exit_code == 0

    assert result.json_output is not None
    assert result.json_output["robotType"] == "OT-2 Standard"
    assert result.json_output["pipettes"] == []
    assert result.json_output["commands"] == []
    assert result.json_output["config"] == {
        "apiVersion": [2, 18],
        "protocolType": "python",
    }
    assert result.json_output["files"] == [{"name": "protocol.py", "role": "main"}]
    [error] = result.json_output["errors"]
    assert error["detail"] == "name 'call_a_non_existent_func' is not defined"
    assert error["errorCode"] == "4000"
    assert error["errorType"] == "UnexpectedAnalysisError"


@pytest.mark.parametrize("output", ["--json-output", "--human-json-output"])
def test_analyze_json_protocol(
    tmp_path: Path,
    output: str,
) -> None:
    """Test that a json protocol analyzes correctly."""
    json_file = (
        Path(__file__).parents[4]
        / "shared-data"
        / "protocol"
        / "fixtures"
        / "8"
        / "simpleV8.json"
    )
    result = _get_analysis_result([json_file], output)

    assert result.exit_code == 0
    op = result.json_output
    assert op is not None
    assert len(op["commands"]) == 27
    assert op["result"] == AnalysisResult.OK.value



from unittest import mock

from opentrons_shared_data.protocol.types import MoveToWellParams

from opentrons.protocol_api import MAX_SUPPORTED_VERSION, InstrumentContext, labware
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_api.core.legacy.legacy_well_core import LegacyWellCore
from opentrons.protocol_api.core.legacy.well_geometry import WellGeometry
from opentrons.protocols.execution.execute_json_v5 import _move_to_well
from opentrons.types import Point


def test_move_to_well_with_optional_params() -> None:
    mock_parent = mock.create_autospec(labware.Labware)
    mock_parent_core = mock.create_autospec(LegacyLabwareCore)
    pipette_mock = mock.create_autospec(InstrumentContext)
    instruments = {"somePipetteId": pipette_mock}

    well = labware.Well(
        parent=mock_parent,
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                {
                    "shape": "circular",
                    "depth": 40,
                    "totalLiquidVolume": 100,
                    "diameter": 30,
                    "x": 40,
                    "y": 50,
                    "z": 3,
                },
                parent_point=Point(10, 20, 30),
                parent_object=mock_parent_core,
            ),
            display_name="some well",
            has_tip=False,
            name="A2",
        ),
        api_version=MAX_SUPPORTED_VERSION,
    )

    mock_get_well = mock.MagicMock(return_value=well, name="mock_get_well")

    params: "MoveToWellParams" = {
        "pipette": "somePipetteId",
        "labware": "someLabwareId",
        "well": "someWell",
        "offset": {"x": 10, "y": 11, "z": 12},
        "forceDirect": mock.sentinel.force_direct,
        "minimumZHeight": mock.sentinel.minimum_z_height,
    }

    with mock.patch(
        "opentrons.protocols.execution.execute_json_v5._get_well", new=mock_get_well
    ):
        _move_to_well(instruments, mock.sentinel.loaded_labware, params)

    assert pipette_mock.mock_calls == [
        mock.call.move_to(
            well.bottom().move(Point(10, 11, 12)),
            force_direct=mock.sentinel.force_direct,
            minimum_z_height=mock.sentinel.minimum_z_height,
        )
    ]

    assert mock_get_well.mock_calls == [mock.call(mock.sentinel.loaded_labware, params)]


def test_move_to_well_without_optional_params() -> None:
    mock_parent = mock.create_autospec(labware.Labware)
    mock_parent_core = mock.create_autospec(LegacyLabwareCore)
    pipette_mock = mock.create_autospec(InstrumentContext)
    instruments = {"somePipetteId": pipette_mock}

    well = labware.Well(
        parent=mock_parent,
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                {
                    "shape": "circular",
                    "depth": 40,
                    "totalLiquidVolume": 100,
                    "diameter": 30,
                    "x": 40,
                    "y": 50,
                    "z": 3,
                },
                parent_point=Point(10, 20, 30),
                parent_object=mock_parent_core,
            ),
            display_name="some well",
            has_tip=False,
            name="A2",
        ),
        api_version=MAX_SUPPORTED_VERSION,
    )

    mock_get_well = mock.MagicMock(return_value=well, name="mock_get_well")

    params: "MoveToWellParams" = {
        "pipette": "somePipetteId",
        "labware": "someLabwareId",
        "well": "someWell",
    }

    with mock.patch(
        "opentrons.protocols.execution.execute_json_v5._get_well", new=mock_get_well
    ):
        _move_to_well(instruments, mock.sentinel.loaded_labware, params)

    assert pipette_mock.mock_calls == [
        mock.call.move_to(well.bottom(), force_direct=False, minimum_z_height=None)
    ]

    assert mock_get_well.mock_calls == [mock.call(mock.sentinel.loaded_labware, params)]



"""Command models to engage a user to fill a Flex Stacker."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Type

from ...errors import (
    ErrorOccurrence,
)
from ...errors.exceptions import FlexStackerLabwarePoolNotYetDefinedError
from ...types import (
    SYSTEM_LOCATION,
    InStackerHopperLocation,
    LabwareLocationSequence,
    NotOnDeckLocationSequenceComponent,
    StackerFillEmptyStrategy,
    StackerStoredLabwareGroup,
)
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from .common import (
    INITIAL_STORED_LABWARE_DESCRIPTION,
    adapter_location_sequences_with_default,
    build_ids_to_fill,
    build_or_assign_labware_to_hopper,
    labware_location_base_sequence,
    labware_locations_for_group,
    lid_location_sequences_with_default,
    primary_location_sequences,
)

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler, RunControlHandler
    from opentrons.protocol_engine.resources import ModelUtils
    from opentrons.protocol_engine.state.state import StateView


FillCommandType = Literal["flexStacker/fill"]


class FillParams(BaseModel):
    """The parameters defining how a stacker should be filled."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")

    strategy: StackerFillEmptyStrategy = Field(
        ...,
        description=(
            "How to fill the stacker. "
            "If manualWithPause, pause the protocol until the client sends an interaction, and apply "
            "the new specified count thereafter. If logical, do not pause but immediately apply the "
            "specified count."
        ),
    )

    message: str | SkipJsonSchema[None] = Field(
        None,
        description="The message to display on connected clients during a manualWithPause strategy fill.",
    )

    count: Optional[Annotated[int, Field(ge=1)]] = Field(
        None,
        description=(
            "How full the labware pool should now be. If None, default to the maximum amount "
            "of the currently-configured labware the pool can hold. "
            "If this number is larger than the maximum the pool can hold, it will be clamped to "
            "the maximum. If this number is smaller than the current amount of labware the pool "
            "holds, it will be clamped to that minimum. Do not use the value in the parameters as "
            "an outside observer; instead, use the count value from the results."
        ),
    )
    labwareToStore: list[StackerStoredLabwareGroup] | None = Field(
        None, description=INITIAL_STORED_LABWARE_DESCRIPTION
    )


class FillResult(BaseModel):
    """Result data from a stacker fill command."""

    count: int = Field(
        ..., description="The new amount of labware stored in the stacker."
    )
    primaryLabwareURI: str = Field(
        ..., description="The labware definition URI of the primary labware."
    )
    adapterLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the adapter labware.",
    )
    lidLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the lid labware.",
    )
    storedLabware: list[StackerStoredLabwareGroup] | SkipJsonSchema[None] = Field(
        None, description="The labware now stored in the stacker."
    )
    addedLabware: list[StackerStoredLabwareGroup] | SkipJsonSchema[None] = Field(
        None, description="The labware just added to the stacker."
    )
    originalPrimaryLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each primary labware, in the same order as storedLabware.",
    )
    originalAdapterLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each adapter labware, in the same order as storedLabware. None if the pool does not specify an adapter.",
    )
    originalLidLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each lid labware, in the same order as storedLabware. None if the pool does not specify a lid.",
    )
    newPrimaryLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each primary labware, in the same order as storedLabware.",
    )
    newAdapterLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each adapter labware, in the same order as storedLabware. None if the pool does not specify an adapter.",
    )
    newLidLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each lid labware, in the same order as storedLabware. None if the pool does not specify a lid labware.",
    )


class FillImpl(AbstractCommandImpl[FillParams, SuccessData[FillResult]]):
    """Implementation of a stacker fill command."""

    def __init__(
        self,
        state_view: StateView,
        run_control: RunControlHandler,
        model_utils: ModelUtils,
        equipment: EquipmentHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._run_control = run_control
        self._model_utils = model_utils
        self._equipment = equipment

    async def execute(self, params: FillParams) -> SuccessData[FillResult]:
        """Execute the stacker fill command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )

        if stacker_state.pool_primary_definition is None:
            location = self._state_view.modules.get_location(params.moduleId)
            raise FlexStackerLabwarePoolNotYetDefinedError(
                message=f"The Flex Stacker in {location} has not been configured yet and cannot be filled."
            )

        groups_to_load = build_ids_to_fill(
            stacker_state.pool_adapter_definition is not None,
            stacker_state.pool_lid_definition is not None,
            params.labwareToStore,
            params.count,
            stacker_state.max_pool_count,
            len(stacker_state.contained_labware_bottom_first),
            self._model_utils,
        )

        state_update, new_contained_labware = await build_or_assign_labware_to_hopper(
            stacker_state.pool_primary_definition,
            stacker_state.pool_adapter_definition,
            stacker_state.pool_lid_definition,
            params.moduleId,
            groups_to_load,
            stacker_state.get_contained_labware(),
            self._equipment,
            self._state_view,
        )
        added_labware = [
            labware
            for labware in new_contained_labware
            if labware not in stacker_state.contained_labware_bottom_first
        ]

        original_location_sequences = [
            labware_locations_for_group(
                group,
                labware_location_base_sequence(
                    group,
                    self._state_view,
                    [
                        NotOnDeckLocationSequenceComponent(
                            logicalLocationName=SYSTEM_LOCATION
                        )
                    ],
                ),
            )
            for group in added_labware
        ]
        new_location_sequences = [
            labware_locations_for_group(
                group, [InStackerHopperLocation(moduleId=params.moduleId)]
            )
            for group in added_labware
        ]

        stacker_hw = self._equipment.get_module_hardware_api(
            module_id=stacker_state.module_id
        )
        if stacker_hw:
            stacker_hw.set_stacker_identify(True)

        if params.strategy == StackerFillEmptyStrategy.MANUAL_WITH_PAUSE:
            await self._run_control.wait_for_resume()

        if stacker_hw:
            stacker_hw.set_stacker_identify(False)

        return SuccessData(
            public=FillResult.model_construct(
                count=len(new_contained_labware),
                storedLabware=new_contained_labware,
                addedLabware=added_labware,
                primaryLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_primary_definition
                ),
                adapterLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_adapter_definition
                ),
                lidLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_lid_definition
                ),
                originalPrimaryLabwareLocationSequences=primary_location_sequences(
                    original_location_sequences
                ),
                originalAdapterLabwareLocationSequences=adapter_location_sequences_with_default(
                    original_location_sequences, stacker_state.pool_adapter_definition
                ),
                originalLidLabwareLocationSequences=lid_location_sequences_with_default(
                    original_location_sequences, stacker_state.pool_lid_definition
                ),
                newPrimaryLabwareLocationSequences=primary_location_sequences(
                    new_location_sequences
                ),
                newAdapterLabwareLocationSequences=adapter_location_sequences_with_default(
                    new_location_sequences, stacker_state.pool_adapter_definition
                ),
                newLidLabwareLocationSequences=lid_location_sequences_with_default(
                    new_location_sequences, stacker_state.pool_lid_definition
                ),
            ),
            state_update=state_update,
        )


class Fill(BaseCommand[FillParams, FillResult, ErrorOccurrence]):
    """A command to fill the Flex Stacker with labware."""

    commandType: FillCommandType = "flexStacker/fill"
    params: FillParams
    result: Optional[FillResult] = None

    _ImplementationCls: Type[FillImpl] = FillImpl


class FillCreate(BaseCommandCreate[FillParams]):
    """A request to execute a Flex Stacker fill command."""

    commandType: FillCommandType = "flexStacker/fill"
    params: FillParams

    _CommandCls: Type[Fill] = Fill



"""Pipetting execution handler."""

from typing import cast
from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import (
    CuboidalFrustum,
    InnerWellGeometry,
    RectangularWellDefinition3,
    SphericalSegment,
)

from ..note_utils import CommandNoteMatcher
from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.types import CriticalPoint
from opentrons.protocol_engine.errors.exceptions import (
    InvalidAspirateVolumeError,
    InvalidDispenseVolumeError,
    InvalidPushOutVolumeError,
    TipNotAttachedError,
)
from opentrons.protocol_engine.execution.pipetting import (
    HardwarePipettingHandler,
    VirtualPipettingHandler,
    create_pipetting_handler,
)
from opentrons.protocol_engine.notes import CommandNote, CommandNoteAdder
from opentrons.protocol_engine.state.labware import LabwareView
from opentrons.protocol_engine.state.pipettes import HardwarePipette
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.wells import WellView
from opentrons.protocol_engine.types import TipGeometry
from opentrons.types import Mount, Point

_TEST_INNER_WELL_GEOMETRY = InnerWellGeometry(
    sections=[
        CuboidalFrustum(
            shape="cuboidal",
            topXDimension=7.6,
            topYDimension=8.5,
            bottomXDimension=5.6,
            bottomYDimension=6.5,
            topHeight=45,
            bottomHeight=20,
        ),
        CuboidalFrustum(
            shape="cuboidal",
            topXDimension=5.6,
            topYDimension=6.5,
            bottomXDimension=4.5,
            bottomYDimension=4.0,
            topHeight=20,
            bottomHeight=10,
        ),
        SphericalSegment(
            shape="spherical",
            radiusOfCurvature=6,
            topHeight=10,
            bottomHeight=0.0,
        ),
    ],
)


@pytest.fixture
def mock_hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mock in the shape of a HardwareAPI."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock in the shape of a StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def hardware_subject(
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
) -> HardwarePipettingHandler:
    """Get a HardwarePipettingHandler test subject."""
    return HardwarePipettingHandler(
        state_view=mock_state_view, hardware_api=mock_hardware_api
    )


@pytest.fixture
def mock_labware_view(decoy: Decoy) -> LabwareView:
    """Get a mock in the shape of a LabwareView."""
    return decoy.mock(cls=LabwareView)


@pytest.fixture
def mock_well_view(decoy: Decoy) -> WellView:
    """Get a mock in the shape of a WellView."""
    return decoy.mock(cls=WellView)


async def test_create_pipette_handler(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
) -> None:
    """It should return virtual or real tip handlers depending on config."""
    decoy.when(mock_state_view.config.use_virtual_pipettes).then_return(False)
    assert isinstance(
        create_pipetting_handler(
            state_view=mock_state_view, hardware_api=mock_hardware_api
        ),
        HardwarePipettingHandler,
    )

    decoy.when(mock_state_view.config.use_virtual_pipettes).then_return(True)
    assert isinstance(
        create_pipetting_handler(
            state_view=mock_state_view, hardware_api=mock_hardware_api
        ),
        VirtualPipettingHandler,
    )


@pytest.mark.parametrize(
    ("aspirated_volume", "ready_to_aspirate", "expected"),
    [
        (0.0, True, True),
        (1.0, True, True),
        (1.0, False, False),
    ],
)
def test_hw_get_is_ready_to_aspirate(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
    aspirated_volume: float,
    ready_to_aspirate: bool,
    expected: bool,
) -> None:
    """It should be ready to aspirate if state and HW agree that we're ready."""
    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        aspirated_volume
    )
    decoy.when(
        mock_state_view.pipettes.get_ready_to_aspirate("pipette-id")
    ).then_return(ready_to_aspirate)
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette("pipette-id", {})
    ).then_return(
        HardwarePipette(
            mount=Mount.RIGHT,
            config=cast(PipetteDict, {"ready_to_aspirate": ready_to_aspirate}),
        )
    )

    assert hardware_subject.get_is_ready_to_aspirate("pipette-id") == expected


def test_hw_get_is_ready_to_aspirate_raises_no_tip_attached(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
) -> None:
    """Should raise a TipNotAttachedError error."""
    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_raise(
        TipNotAttachedError()
    )
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette("pipette-id", {})
    ).then_return(
        HardwarePipette(
            mount=Mount.RIGHT,
            config=cast(PipetteDict, {"ready_to_aspirate": True}),
        )
    )

    with pytest.raises(TipNotAttachedError):
        assert hardware_subject.get_is_ready_to_aspirate("pipette-id")


async def test_hw_dispense_in_place(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
) -> None:
    """It should find the pipette by ID and use it to dispense."""
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        25
    )

    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.RIGHT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    result = await hardware_subject.dispense_in_place(
        pipette_id="pipette-id",
        volume=25,
        flow_rate=2.5,
        push_out=None,
        is_full_dispense=True,
    )

    assert result == 25

    decoy.verify(
        mock_hardware_api.set_flow_rate(
            mount=Mount.RIGHT, aspirate=None, dispense=2.5, blow_out=None
        ),
        await mock_hardware_api.dispense(
            mount=Mount.RIGHT,
            volume=25,
            push_out=None,
            correction_volume=0,
            is_full_dispense=True,
        ),
        mock_hardware_api.set_flow_rate(
            mount=Mount.RIGHT, aspirate=1.23, dispense=4.56, blow_out=7.89
        ),
    )


async def test_hw_dispense_in_place_raises_invalid_push_out(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
) -> None:
    """It should raise an InvalidPushOutVolumeError."""
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        25
    )

    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.RIGHT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    with pytest.raises(InvalidPushOutVolumeError):
        await hardware_subject.dispense_in_place(
            pipette_id="pipette-id",
            volume=25,
            flow_rate=2.5,
            push_out=-7,
            is_full_dispense=True,
        )


async def test_hw_dispense_while_tracking(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_view: LabwareView,
    mock_well_view: WellView,
    hardware_subject: HardwarePipettingHandler,
) -> None:
    """Should set flow_rate and call hardware_api aspirate."""
    decoy.when(mock_labware_view.get_well_definition("labware-id", "A1")).then_return(
        RectangularWellDefinition3.model_construct(totalLiquidVolume=1100000)  # type: ignore[call-arg]
    )
    decoy.when(mock_labware_view.get_well_geometry("labware-id", "A1")).then_return(
        _TEST_INNER_WELL_GEOMETRY
    )

    decoy.when(mock_state_view.pipettes.get_working_volume("pipette-id")).then_return(
        25
    )
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        25
    )

    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.LEFT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    decoy.when(
        mock_state_view.geometry.get_liquid_handling_z_change(
            labware_id="labware-id",
            well_name="A1",
            pipette_id="pipette-id",
            operation_volume=25.0,
        )
    ).then_return(4.544)

    decoy.when(
        mock_state_view.motion.get_critical_point_for_wells_in_labware("labware-id")
    ).then_return(CriticalPoint.XY_CENTER)

    result = await hardware_subject.dispense_while_tracking(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="A1",
        volume=25,
        flow_rate=2.5,
        end_point=Point(0, 0, 0),
        push_out=2.0,
        is_full_dispense=True,
        movement_delay=3.0,
    )
    # make sure hw dispense_while_tracking runs without error
    assert result == 25
    decoy.verify(
        await mock_hardware_api.dispense_while_tracking(
            mount=Mount.LEFT,
            end_point=Point(0, 0, 0),
            volume=25,
            end_critical_point=CriticalPoint.XY_CENTER,
            push_out=2.0,
            is_full_dispense=True,
            movement_delay=3.0,
        )
    )


async def test_hw_aspirate_while_tracking(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_view: LabwareView,
    mock_well_view: WellView,
    hardware_subject: HardwarePipettingHandler,
    mock_command_note_adder: CommandNoteAdder,
) -> None:
    """Should set flow_rate and call hardware_api aspirate."""
    decoy.when(mock_labware_view.get_well_definition("labware-id", "A1")).then_return(
        RectangularWellDefinition3.model_construct(totalLiquidVolume=1100000)  # type: ignore[call-arg]
    )
    decoy.when(mock_labware_view.get_well_geometry("labware-id", "A1")).then_return(
        _TEST_INNER_WELL_GEOMETRY
    )

    decoy.when(mock_state_view.pipettes.get_working_volume("pipette-id")).then_return(
        25
    )
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        0
    )

    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.LEFT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    decoy.when(
        mock_state_view.geometry.get_liquid_handling_z_change(
            labware_id="labware-id",
            well_name="A1",
            pipette_id="pipette-id",
            operation_volume=-25.0,
        )
    ).then_return(4.544)

    decoy.when(
        mock_state_view.motion.get_critical_point_for_wells_in_labware("labware-id")
    ).then_return(CriticalPoint.Y_CENTER)

    result = await hardware_subject.aspirate_while_tracking(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="A1",
        volume=25,
        flow_rate=2.5,
        end_point=Point(0, 0, 0),
        command_note_adder=mock_command_note_adder,
        movement_delay=3.0,
    )
    # make sure hw aspirate_while_tracking runs without error
    assert result == 25
    decoy.verify(
        await mock_hardware_api.aspirate_while_tracking(
            mount=Mount.LEFT,
            end_point=Point(0, 0, 0),
            volume=25,
            movement_delay=3.0,
            end_critical_point=CriticalPoint.Y_CENTER,
        )
    )


async def test_hw_aspirate_in_place(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
    mock_command_note_adder: CommandNoteAdder,
) -> None:
    """Should set flow_rate and call hardware_api aspirate."""
    decoy.when(mock_state_view.pipettes.get_working_volume("pipette-id")).then_return(
        25
    )
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        0
    )

    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.LEFT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    result = await hardware_subject.aspirate_in_place(
        pipette_id="pipette-id",
        volume=25,
        flow_rate=2.5,
        command_note_adder=mock_command_note_adder,
    )

    assert result == 25

    decoy.verify(
        mock_hardware_api.set_flow_rate(
            mount=Mount.LEFT, aspirate=2.5, dispense=None, blow_out=None
        ),
        await mock_hardware_api.aspirate(
            mount=Mount.LEFT, volume=25, correction_volume=0
        ),
        mock_hardware_api.set_flow_rate(
            mount=Mount.LEFT, aspirate=1.23, dispense=4.56, blow_out=7.89
        ),
    )


async def test_virtual_blow_out_in_place(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
) -> None:
    """Should set flow_rate and call hardware_api blow-out."""
    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.LEFT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    await hardware_subject.blow_out_in_place(pipette_id="pipette-id", flow_rate=2.5)

    decoy.verify(
        mock_hardware_api.set_flow_rate(
            mount=Mount.LEFT, aspirate=None, dispense=None, blow_out=2.5
        ),
        await mock_hardware_api.blow_out(mount=Mount.LEFT),
        mock_hardware_api.set_flow_rate(
            mount=Mount.LEFT, aspirate=1.23, dispense=4.56, blow_out=7.89
        ),
    )


def test_virtual_get_is_ready_to_aspirate(
    decoy: Decoy, mock_state_view: StateView
) -> None:
    """Should check if pipette is ready to aspirate."""
    subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(
        mock_state_view.pipettes.get_aspirated_volume(pipette_id="pipette-id")
    ).then_raise(TipNotAttachedError())

    decoy.when(
        mock_state_view.pipettes.get_ready_to_aspirate(pipette_id="pipette-id")
    ).then_return(True)

    with pytest.raises(TipNotAttachedError):
        subject.get_is_ready_to_aspirate(pipette_id="pipette-id")

    decoy.when(
        mock_state_view.pipettes.get_aspirated_volume(pipette_id="pipette-id-123")
    ).then_return(0)

    decoy.when(
        mock_state_view.pipettes.get_ready_to_aspirate(pipette_id="pipette-id-123")
    ).then_return(True)

    assert subject.get_is_ready_to_aspirate(pipette_id="pipette-id-123") is True

    decoy.when(
        mock_state_view.pipettes.get_ready_to_aspirate(pipette_id="pipette-id-123")
    ).then_return(False)

    assert subject.get_is_ready_to_aspirate(pipette_id="pipette-id-123") is False


async def test_virtual_aspirate_in_place(
    mock_state_view: StateView, decoy: Decoy, mock_command_note_adder: CommandNoteAdder
) -> None:
    """Should return the volume."""
    decoy.when(
        mock_state_view.pipettes.get_working_volume(pipette_id="pipette-id")
    ).then_return(3)

    decoy.when(
        mock_state_view.pipettes.get_aspirated_volume(pipette_id="pipette-id")
    ).then_return(1)

    subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=1, diameter=2, volume=3)
    )

    result = await subject.aspirate_in_place(
        pipette_id="pipette-id",
        volume=2,
        flow_rate=5,
        command_note_adder=mock_command_note_adder,
    )
    assert result == 2


async def test_virtual_dispense_in_place(
    decoy: Decoy, mock_state_view: StateView
) -> None:
    """Should return the volume."""
    subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=1, diameter=2, volume=3)
    )

    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        3
    )

    result = await subject.dispense_in_place(
        pipette_id="pipette-id",
        volume=3,
        flow_rate=5,
        push_out=None,
        is_full_dispense=True,
    )
    assert result == 3


async def test_virtual_dispense_in_place_raises_invalid_push_out(
    decoy: Decoy, mock_state_view: StateView
) -> None:
    """Should raise an InvalidPushOutVolumeError."""
    subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=1, diameter=2, volume=3)
    )

    with pytest.raises(InvalidPushOutVolumeError):
        await subject.dispense_in_place(
            pipette_id="pipette-id",
            volume=3,
            flow_rate=5,
            push_out=-7,
            is_full_dispense=False,
        )


async def test_virtual_dispense_in_place_raises_invalid_volume(
    decoy: Decoy, mock_state_view: StateView
) -> None:
    """Should raise an InvalidDispenseVolumeError."""
    subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=1, diameter=2, volume=3)
    )

    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        None
    )

    with pytest.raises(InvalidDispenseVolumeError):
        await subject.dispense_in_place(
            pipette_id="pipette-id",
            volume=3,
            flow_rate=5,
            push_out=7,
            is_full_dispense=False,
        )


async def test_virtual_aspirate_validate_tip_attached(
    mock_state_view: StateView, decoy: Decoy, mock_command_note_adder: CommandNoteAdder
) -> None:
    """Should raise an error that a tip is not attached."""
    subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        None
    )

    with pytest.raises(
        TipNotAttachedError, match="Cannot perform aspirate without a tip attached"
    ):
        await subject.aspirate_in_place(
            "pipette-id",
            volume=20,
            flow_rate=1,
            command_note_adder=mock_command_note_adder,
        )


async def test_virtual_dispense_validate_tip_attached(
    mock_state_view: StateView, decoy: Decoy
) -> None:
    """Should raise an error that a tip is not attached."""
    subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        None
    )

    with pytest.raises(
        TipNotAttachedError, match="Cannot perform dispense without a tip attached"
    ):
        await subject.dispense_in_place(
            "pipette-id", volume=20, flow_rate=1, push_out=None, is_full_dispense=False
        )


async def test_aspirate_volume_validation(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
    mock_command_note_adder: CommandNoteAdder,
) -> None:
    """It should validate the input volume, possibly adjusting it for rounding error.

    This is done on both the VirtualPipettingHandler and HardwarePipettingHandler
    because they should behave the same way.
    """
    virtual_subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=1, diameter=2, volume=3)
    )
    decoy.when(mock_state_view.pipettes.get_working_volume("pipette-id")).then_return(3)
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        2
    )

    # Stuff that only matters for the hardware subject:
    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.LEFT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    ok_volume = 1.0000000000001
    not_ok_volume = 1.01
    expected_adjusted_volume = 1

    for subject in [virtual_subject, hardware_subject]:
        assert (
            await subject.aspirate_in_place(
                pipette_id="pipette-id",
                volume=ok_volume,
                flow_rate=1,
                command_note_adder=mock_command_note_adder,
            )
            == expected_adjusted_volume
        )
        decoy.verify(
            mock_command_note_adder(
                cast(
                    CommandNote,
                    CommandNoteMatcher(
                        matching_noteKind_regex="warning",
                        matching_shortMessage_regex="Aspirate clamped to 1 µL",
                    ),
                )
            )
        )
        with pytest.raises(InvalidAspirateVolumeError):
            await subject.aspirate_in_place(
                pipette_id="pipette-id",
                volume=not_ok_volume,
                flow_rate=1,
                command_note_adder=mock_command_note_adder,
            )


async def test_dispense_volume_validation(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
) -> None:
    """It should validate the input volume, possibly adjusting it for rounding error.

    This is done on both the VirtualPipettingHandler and HardwarePipettingHandler
    because they should behave the same way.
    """
    virtual_subject = VirtualPipettingHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        TipGeometry(length=1, diameter=2, volume=3)
    )
    decoy.when(mock_state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(
        1
    )

    # Stuff that only matters for the hardware subject:
    decoy.when(mock_hardware_api.attached_instruments).then_return({})
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes={},
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.LEFT,
            config=cast(
                PipetteDict,
                {
                    "aspirate_flow_rate": 1.23,
                    "dispense_flow_rate": 4.56,
                    "blow_out_flow_rate": 7.89,
                },
            ),
        )
    )

    ok_volume = 1.0000000000001
    not_ok_volume = 1.01
    expected_adjusted_volume = 1

    for subject in [virtual_subject, hardware_subject]:
        assert (
            await subject.dispense_in_place(
                pipette_id="pipette-id",
                volume=ok_volume,
                flow_rate=5,
                push_out=7,
                is_full_dispense=True,
            )
            == expected_adjusted_volume
        )
        with pytest.raises(InvalidDispenseVolumeError):
            await subject.dispense_in_place(
                pipette_id="pipette-id",
                volume=not_ok_volume,
                flow_rate=5,
                push_out=7,
                is_full_dispense=True,
            )


async def test_hw_increase_evo_disp_count(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    hardware_subject: HardwarePipettingHandler,
) -> None:
    """Should set flow_rate and call hardware_api aspirate."""
    decoy.when(mock_hardware_api.attached_instruments).then_return(
        sentinel.attached_instruments
    )
    decoy.when(
        mock_state_view.pipettes.get_hardware_pipette(
            pipette_id="pipette-id", attached_pipettes=sentinel.attached_instruments
        )
    ).then_return(
        HardwarePipette(
            mount=Mount.LEFT,
            config=cast(
                PipetteDict,
                {},
            ),
        )
    )
    await hardware_subject.increase_evo_disp_count(pipette_id="pipette-id")

    decoy.verify(await mock_hardware_api.increase_evo_disp_count(mount=Mount.LEFT))



from typing import TYPE_CHECKING, Any

import pytest

from opentrons_shared_data.pipette.types import LabwareUri

from opentrons import config
from opentrons.calibration_storage import (
    file_operators as io,
)
from opentrons.calibration_storage import (
    helpers,
)
from opentrons.calibration_storage import (
    types as cs_types,
)
from opentrons.calibration_storage.ot2 import (
    clear_tip_length_calibration,
    create_tip_length_data,
    delete_tip_length_calibration,
    load_tip_length_calibration,
    models,
    save_tip_length_calibration,
    tip_lengths_for_pipette,
)

if TYPE_CHECKING:
    from opentrons_shared_data.labware.types import LabwareDefinition2


@pytest.fixture
def starting_calibration_data(
    ot_config_tempdir: Any,
    minimal_labware_def: "LabwareDefinition2",
    minimal_labware_def2: "LabwareDefinition2",
) -> None:
    """
    Starting calibration data fixture.

    Adds dummy data to a temporary directory to test delete commands against.
    """
    tip_length1 = create_tip_length_data(minimal_labware_def, 22.0)
    tip_length2 = create_tip_length_data(minimal_labware_def, 31.0)
    tip_length3 = create_tip_length_data(minimal_labware_def2, 31.0)
    save_tip_length_calibration("pip1", tip_length1)
    save_tip_length_calibration("pip2", tip_length2)
    save_tip_length_calibration("pip1", tip_length3)
    inside_data = tip_length3[LabwareUri("dummy_namespace/minimal_labware_def/1")]
    data = {
        inside_data.definitionHash: {
            "tipLength": 27,
            "lastModified": inside_data.lastModified.isoformat(),
            "source": inside_data.source,
            "status": inside_data.status.model_dump(),
            "uri": "dummy_namespace/minimal_labware_def/1",
        }
    }
    tip_length_dir_path = config.get_tip_length_cal_path()
    io.save_to_file(tip_length_dir_path, "pip2", data)


def test_save_tip_length_calibration(
    ot_config_tempdir: Any, minimal_labware_def: "LabwareDefinition2"
) -> None:
    """
    Test saving tip length calibrations.
    """
    assert tip_lengths_for_pipette("pip1") == {}
    assert tip_lengths_for_pipette("pip2") == {}
    tip_rack_uri = helpers.uri_from_definition(minimal_labware_def)
    tip_length1 = create_tip_length_data(minimal_labware_def, 22.0)
    tip_length2 = create_tip_length_data(minimal_labware_def, 31.0)
    save_tip_length_calibration("pip1", tip_length1)
    save_tip_length_calibration("pip2", tip_length2)
    assert tip_lengths_for_pipette("pip1")[tip_rack_uri].tipLength == 22.0
    assert tip_lengths_for_pipette("pip2")[tip_rack_uri].tipLength == 31.0


def test_get_tip_length_calibration(
    starting_calibration_data: Any, minimal_labware_def: "LabwareDefinition2"
) -> None:
    """
    Test ability to get a tip length calibration model.
    """
    tip_length_data = load_tip_length_calibration("pip1", minimal_labware_def)
    tip_rack_hash = helpers.hash_labware_def(minimal_labware_def)
    assert tip_length_data == models.v1.TipLengthModel(
        tipLength=22.0,
        source=cs_types.SourceType.user,
        lastModified=tip_length_data.lastModified,
        definitionHash=tip_rack_hash,
    )

    with pytest.raises(cs_types.TipLengthCalNotFound):
        load_tip_length_calibration("nopipette", minimal_labware_def)


def test_delete_specific_tip_calibration(
    starting_calibration_data: Any, minimal_labware_def: "LabwareDefinition2"
) -> None:
    """
    Test delete a specific tip length calibration.
    """
    assert len(tip_lengths_for_pipette("pip1").keys()) == 2
    assert tip_lengths_for_pipette("pip2") != {}
    tip_rack_uri = helpers.uri_from_definition(minimal_labware_def)
    delete_tip_length_calibration("pip1", tiprack_uri=tip_rack_uri)
    assert len(tip_lengths_for_pipette("pip1").keys()) == 1
    assert tip_lengths_for_pipette("pip2") != {}


def test_delete_all_tip_calibration(starting_calibration_data: Any) -> None:
    """
    Test delete all tip length calibration.
    """
    assert tip_lengths_for_pipette("pip1") != {}
    assert tip_lengths_for_pipette("pip2") != {}
    clear_tip_length_calibration()
    assert tip_lengths_for_pipette("pip1") == {}
    assert tip_lengths_for_pipette("pip2") == {}


def test_uriless_calibrations_are_dropped(ot_config_tempdir: object) -> None:
    """Legacy records without a `uri` field should be silently ignored."""

    data = {
        "ed323db6ca1ddf197aeb20667c1a7a91c89cfb2f931f45079d483928da056812": {
            "tipLength": 123,
            "lastModified": "2021-01-11T00:34:29.291073+00:00",
            "source": "user",
            "status": {"markedBad": False},
        },
        "130e17bb7b2f0c0472dcc01c1ff6f600ca1a6f9f86a90982df56c4bf43776824": {
            "tipLength": 456,
            "lastModified": "2021-05-12T22:16:14.249567+00:00",
            "source": "user",
            "status": {"markedBad": False},
            "uri": "opentrons/opentrons_96_filtertiprack_200ul/1",
        },
    }

    io.save_to_file(config.get_tip_length_cal_path(), "pipette1234", data)
    result = tip_lengths_for_pipette("pipette1234")
    assert len(result) == 1
    assert (
        result[LabwareUri("opentrons/opentrons_96_filtertiprack_200ul/1")].tipLength
        == 456
    )



"""Command models to pulse the light on a module for identification."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional

from pydantic import BaseModel, Field
from typing_extensions import Type

from ..errors.error_occurrence import ErrorOccurrence
from ..types import ModuleModel
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData

if TYPE_CHECKING:
    from ..state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

IdentifyModuleCommandType = Literal["identifyModule"]


class IdentifyModuleParams(BaseModel):
    """The parameters defining the module to be identified."""

    model: ModuleModel = Field(..., description="The model of the module")
    moduleId: str = Field(..., description="Unique ID of the module")
    start: bool = Field(..., description="Start or stop identify")
    color: Optional[str] = Field(None, description="Optional color to identify module")


class IdentifyModuleResult(BaseModel):
    """Result data from an IdentifyModule command."""


class IdentifyModuleImpl(
    AbstractCommandImpl[IdentifyModuleParams, SuccessData[IdentifyModuleResult]]
):
    """Implementation of an IdentifyModule command."""

    def __init__(
        self, state_view: StateView, equipment: EquipmentHandler, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: IdentifyModuleParams
    ) -> SuccessData[IdentifyModuleResult]:
        """Execute the IdentifyModule command."""
        # ONLY flex stacker has support for identify module command...for now
        if params.model == ModuleModel.FLEX_STACKER_MODULE_V1:
            module_substate = self._state_view.modules.get_flex_stacker_substate(
                module_id=params.moduleId
            )
            module_hw = self._equipment.get_module_hardware_api(
                module_substate.module_id
            )
            if module_hw is not None:
                await module_hw.identify(params.start, params.color)
        else:
            raise NotImplementedError(
                f"IdentifyModule is not supported for {params.model}"
            )

        return SuccessData(public=IdentifyModuleResult())


class IdentifyModule(
    BaseCommand[IdentifyModuleParams, IdentifyModuleResult, ErrorOccurrence]
):
    """A command to identify a module."""

    commandType: IdentifyModuleCommandType = "identifyModule"
    params: IdentifyModuleParams
    result: Optional[IdentifyModuleResult] = None

    _ImplementationCls: Type[IdentifyModuleImpl] = IdentifyModuleImpl


class IdentifyModuleCreate(BaseCommandCreate[IdentifyModuleParams]):
    """A request to execute an IdentifyModule command."""

    commandType: IdentifyModuleCommandType = "identifyModule"
    params: IdentifyModuleParams

    _CommandCls: Type[IdentifyModule] = IdentifyModule



"""Test instrument context simulation."""

from typing import cast

import pytest
from _pytest.fixtures import SubRequest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.protocol_api.core.common import LabwareCore, ProtocolCore
from opentrons.types import Location, Mount


@pytest.fixture(
    params=[
        lazy_fixture("protocol_context"),
        lazy_fixture("simulating_protocol_context"),
    ]
)
def subject(request: SubRequest) -> ProtocolCore:
    return cast(ProtocolCore, request.param)


@pytest.mark.ot2_only
def test_replacing_instrument_tip_state(
    subject: ProtocolCore, tip_rack: LabwareCore
) -> None:
    """It should refer to same state when replacing the same pipette."""
    # This test validates that bug https://github.com/Opentrons/opentrons/issues/8273
    # is fixed. The user who found this calls load_instrument repeatedly with same name,
    # mount, and replace=True. The resulting InstrumentContext references are used
    # interchangeably causing fast simulation to fail due to the pipette state being
    # attached to the InstrumentContextSimulation object.
    # The solution is to reuse InstrumentContextSimulation instances when the user is
    # replacing the same pipette at the same mount.
    subject.home()
    pip1 = subject.load_instrument(PipetteNameType.P300_SINGLE_GEN2, Mount.RIGHT)
    pip2 = subject.load_instrument(PipetteNameType.P300_SINGLE_GEN2, Mount.RIGHT)

    pip1.pick_up_tip(
        location=Location(
            point=tip_rack.get_well_core("A1").get_top(z_offset=0), labware=None
        ),
        well_core=tip_rack.get_well_core("A1"),
        presses=None,
        increment=None,
        prep_after=False,
    )
    assert pip1.has_tip() is True
    assert pip2.has_tip() is True

    pip2.drop_tip(
        location=None, well_core=tip_rack.get_well_core("A1"), home_after=False
    )

    assert pip1.has_tip() is False
    assert pip2.has_tip() is False


@pytest.mark.ot2_only
def test_load_instrument_raises(simulating_protocol_context: ProtocolCore) -> None:
    with pytest.raises(ValueError):
        simulating_protocol_context.load_instrument(
            instrument_name=PipetteNameType.P1000_SINGLE_FLEX, mount=Mount.RIGHT
        )



from opentrons_shared_data.util import StrEnum


class ModuleType(StrEnum):
    """Module type enumeration."""

    Magnetic = "magdeck"
    Temperature = "tempdeck"
    Thermocycler = "thermocycler"
    Heatershaker = "heatershaker"



"""Integration tests for the Python Protocol API.

These test the Python Protocol API from the point of view of the user's Python protocol.

They do not make sure the robot would actually move to the right place--that is the job of
other tests, such as g-code-testing.

These are supplementary for Python Protocol API features that depend on nontrivial interaction
between those layers.
"""



"""Tests for the legacy Protocol API module core implementations."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import TempDeck
from opentrons.hardware_control.modules.types import (
    TemperatureModuleModel,
    TemperatureStatus,
)
from opentrons.protocol_api.core.legacy.legacy_module_core import (
    LegacyTemperatureModuleCore,
    create_module_core,
)
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocol_api.core.legacy.module_geometry import ModuleGeometry


@pytest.fixture
def mock_geometry(decoy: Decoy) -> ModuleGeometry:
    """Get a mock module geometry."""
    return decoy.mock(cls=ModuleGeometry)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SynchronousAdapter[TempDeck]:
    """Get a mock synchronous temperature module hardware."""
    return decoy.mock(name="SynchronousAdapter[TempDeck]")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> LegacyProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=LegacyProtocolCore)


@pytest.fixture
def subject(
    mock_geometry: ModuleGeometry,
    mock_sync_module_hardware: SynchronousAdapter[TempDeck],
    mock_protocol_core: LegacyProtocolCore,
) -> LegacyTemperatureModuleCore:
    """Get a legacy module implementation core with mocked out dependencies."""
    return LegacyTemperatureModuleCore(
        requested_model=TemperatureModuleModel.TEMPERATURE_V1,
        geometry=mock_geometry,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_geometry: ModuleGeometry,
    mock_protocol_core: LegacyProtocolCore,
) -> None:
    """It should be able to create a temperature module core."""
    mock_module_hardware_api = decoy.mock(cls=TempDeck)
    result = create_module_core(
        geometry=mock_geometry,
        module_hardware_api=mock_module_hardware_api,
        requested_model=TemperatureModuleModel.TEMPERATURE_V1,
        protocol_core=mock_protocol_core,
    )

    assert isinstance(result, LegacyTemperatureModuleCore)


def test_set_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[TempDeck],
    subject: LegacyTemperatureModuleCore,
) -> None:
    """It should set the target temperature with the hardware."""
    subject.set_target_temperature(42.0)

    decoy.verify(mock_sync_module_hardware.start_set_temperature(42.0), times=1)


def test_wait_for_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[TempDeck],
    subject: LegacyTemperatureModuleCore,
) -> None:
    """It should wait for the target to be reached with the hardware."""
    subject.wait_for_target_temperature(42.0)

    decoy.verify(mock_sync_module_hardware.await_temperature(42.0), times=1)


def test_deactivate(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[TempDeck],
    subject: LegacyTemperatureModuleCore,
) -> None:
    """It should deactivate the hardware."""
    subject.deactivate()

    decoy.verify(mock_sync_module_hardware.deactivate(), times=1)


def test_get_current_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[TempDeck],
    subject: LegacyTemperatureModuleCore,
) -> None:
    """It should get the current temperature from the hardware."""
    decoy.when(mock_sync_module_hardware.temperature).then_return(42.0)

    result = subject.get_current_temperature()

    assert result == 42.0


def test_get_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[TempDeck],
    subject: LegacyTemperatureModuleCore,
) -> None:
    """It should get the current temperature from the hardware."""
    decoy.when(mock_sync_module_hardware.target).then_return(42.0)

    result = subject.get_target_temperature()

    assert result == 42.0


def test_get_status(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[TempDeck],
    subject: LegacyTemperatureModuleCore,
) -> None:
    """It should get the status string from the hardware."""
    decoy.when(mock_sync_module_hardware.status).then_return(TemperatureStatus.HEATING)

    result = subject.get_status()

    assert result == TemperatureStatus.HEATING



"""Test dispense-in-place commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import PipetteOverpressureError

from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.dispense_while_tracking import (
    DispenseWhileTrackingImplementation,
    DispenseWhileTrackingParams,
    DispenseWhileTrackingResult,
)
from opentrons.protocol_engine.commands.pipetting_common import OverpressureError
from opentrons.protocol_engine.execution import (
    GantryMover,
    MovementHandler,
    PipettingHandler,
)
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import (
    CurrentAddressableArea,
    CurrentPipetteLocation,
    CurrentWell,
    DeckPoint,
    LabwareWellId,
    LiquidHandlingWellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.types import Point


@pytest.fixture
def subject(
    pipetting: PipettingHandler,
    state_view: StateView,
    gantry_mover: GantryMover,
    movement: MovementHandler,
    model_utils: ModelUtils,
) -> DispenseWhileTrackingImplementation:
    """Build a command implementation."""
    return DispenseWhileTrackingImplementation(
        pipetting=pipetting,
        state_view=state_view,
        gantry_mover=gantry_mover,
        model_utils=model_utils,
        movement=movement,
    )


@pytest.mark.parametrize(
    "location,stateupdateLabware,stateupdateWell",
    [
        (
            CurrentWell(
                pipette_id="pipette-id-abc",
                labware_id="funky-labware",
                well_name="funky-well",
            ),
            "funky-labware",
            "funky-well",
        ),
        (
            CurrentAddressableArea("pipette-id-abc", "addressable-area-1"),
            "funky-labware",
            "funky-well",
        ),
        (
            CurrentAddressableArea("pipette-id-abc", "addressable-area-1"),
            "funky-labware",
            "funky-well",
        ),
    ],
)
async def test_dispense_while_tracking_implementation(
    decoy: Decoy,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    movement: MovementHandler,
    state_view: StateView,
    subject: DispenseWhileTrackingImplementation,
    location: CurrentPipetteLocation | None,
    stateupdateLabware: str,
    stateupdateWell: str,
) -> None:
    """It should dispense in place."""
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.MENISCUS,
        offset=WellOffset(x=0, y=0, z=1),
    )
    data = DispenseWhileTrackingParams(
        pipetteId="pipette-id-abc",
        labwareId=stateupdateLabware,
        wellName=stateupdateWell,
        trackFromLocation=well_location,
        trackToLocation=well_location,
        volume=123,
        flowRate=456,
    )
    decoy.when(
        state_view.geometry.get_well_position(
            labware_id=stateupdateLabware,
            well_name=stateupdateWell,
            well_location=well_location,
            operation_volume=123,
            pipette_id="pipette-id-abc",
        )
    ).then_return(Point(1, 2, 3))

    decoy.when(
        await pipetting.dispense_while_tracking(
            pipette_id="pipette-id-abc",
            volume=123,
            flow_rate=456,
            end_point=Point(1, 2, 3),
            push_out=None,
            well_name=stateupdateWell,
            labware_id=stateupdateLabware,
            is_full_dispense=True,
        )
    ).then_return(42)
    decoy.when(state_view.pipettes.get_aspirated_volume("pipette-id-abc")).then_return(
        123
    )

    decoy.when(pipetting.get_state_view()).then_return(state_view)

    decoy.when(state_view.pipettes.get_current_location()).then_return(location)
    decoy.when(
        state_view.pipettes.get_liquid_dispensed_by_ejecting_volume(
            pipette_id="pipette-id-abc", volume=42
        )
    ).then_return(34)

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=stateupdateLabware,
            target_well_name=stateupdateWell,
            pipette_id="pipette-id-abc",
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            stateupdateLabware, stateupdateWell, "pipette-id-abc"
        )
    ).then_return(["A3", "A4"])
    decoy.when(await gantry_mover.get_position("pipette-id-abc")).then_return(
        Point(1, 2, 3)
    )
    _well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.MENISCUS, offset=WellOffset(x=0.0, y=0.0, z=1.0)
    )
    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id-abc",
            labware_id="funky-labware",
            well_name="funky-well",
            well_location=_well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point(x=4, y=5, z=6))

    result = await subject.execute(data)

    if isinstance(location, CurrentWell):
        assert result == SuccessData(
            public=DispenseWhileTrackingResult(
                volume=42,
                position=DeckPoint(x=1, y=2, z=3),
            ),
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                    pipette_id="pipette-id-abc", volume=42
                ),
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id=stateupdateLabware,
                    well_names=["A3", "A4"],
                    volume_added=84,
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id-abc", ready_to_aspirate=False
                ),
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id-abc",
                    new_location=LabwareWellId(
                        labware_id="funky-labware", well_name="funky-well"
                    ),
                    new_deck_point=DeckPoint(x=4, y=5, z=6),
                ),
            ),
        )
    else:
        assert result == SuccessData(
            public=DispenseWhileTrackingResult(
                volume=42,
                position=DeckPoint(x=1, y=2, z=3),
            ),
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                    pipette_id="pipette-id-abc", volume=42, type="ejected"
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id-abc", ready_to_aspirate=False
                ),
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id="funky-labware",
                    well_names=["A3", "A4"],
                    volume_added=84.0,
                ),
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id-abc",
                    new_location=LabwareWellId(
                        labware_id="funky-labware", well_name="funky-well"
                    ),
                    new_deck_point=DeckPoint(x=4, y=5, z=6),
                ),
            ),
        )


@pytest.mark.parametrize(
    "location,stateupdateLabware,stateupdateWell",
    [
        (
            CurrentWell(
                pipette_id="pipette-id",
                labware_id="funky-labware",
                well_name="funky-well",
            ),
            "funky-labware",
            "funky-well",
        ),
        (
            CurrentAddressableArea("pipette-id", "addressable-area-1"),
            "funky-labware",
            "funky-well",
        ),
    ],
)
async def test_overpressure_error(
    decoy: Decoy,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    movement: MovementHandler,
    state_view: StateView,
    model_utils: ModelUtils,
    subject: DispenseWhileTrackingImplementation,
    location: CurrentPipetteLocation | None,
    stateupdateLabware: str,
    stateupdateWell: str,
) -> None:
    """It should return an overpressure error if the hardware API indicates that."""
    pipette_id = "pipette-id"

    position = Point(x=1, y=2, z=3)

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.MENISCUS,
        offset=WellOffset(x=0, y=0, z=1),
    )
    data = DispenseWhileTrackingParams(
        pipetteId=pipette_id,
        labwareId=stateupdateLabware,
        wellName=stateupdateWell,
        trackFromLocation=well_location,
        trackToLocation=well_location,
        volume=50,
        flowRate=1.23,
        pushOut=10,
    )

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=stateupdateLabware,
            target_well_name=stateupdateWell,
            pipette_id="pipette-id",
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            stateupdateLabware, stateupdateWell, "pipette-id"
        )
    ).then_return(["A3", "A4"])
    decoy.when(state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(50)

    decoy.when(pipetting.get_state_view()).then_return(state_view)

    decoy.when(
        state_view.geometry.get_well_position(
            labware_id=stateupdateLabware,
            well_name=stateupdateWell,
            well_location=well_location,
            operation_volume=50,
            pipette_id="pipette-id",
        )
    ).then_return(Point(0, 0, 0))

    decoy.when(
        await pipetting.dispense_while_tracking(
            pipette_id=pipette_id,
            volume=50,
            flow_rate=1.23,
            end_point=Point(0, 0, 0),
            push_out=10,
            well_name=stateupdateWell,
            labware_id=stateupdateLabware,
            is_full_dispense=True,
        ),
    ).then_raise(PipetteOverpressureError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)
    decoy.when(await gantry_mover.get_position(pipette_id)).then_return(position)
    decoy.when(state_view.pipettes.get_current_location()).then_return(location)

    _well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.MENISCUS, offset=WellOffset(x=0.0, y=0.0, z=1.0)
    )
    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="funky-labware",
            well_name="funky-well",
            well_location=_well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point(x=4, y=5, z=6))

    result = await subject.execute(data)

    if isinstance(location, CurrentWell):
        assert result == DefinedErrorData(
            public=OverpressureError.model_construct(
                id=error_id,
                createdAt=error_timestamp,
                wrappedErrors=[matchers.Anything()],
                errorInfo={"retryLocation": (4.0, 5.0, 6.0)},
            ),
            state_update=update_types.StateUpdate(
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id=stateupdateLabware,
                    well_names=["A3", "A4"],
                    volume_added=update_types.CLEAR,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id", ready_to_aspirate=False
                ),
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id",
                    new_location=LabwareWellId(
                        labware_id="funky-labware", well_name="funky-well"
                    ),
                    new_deck_point=DeckPoint(x=4, y=5, z=6),
                ),
            ),
        )
    else:
        assert result == DefinedErrorData(
            public=OverpressureError.model_construct(
                id=error_id,
                createdAt=error_timestamp,
                wrappedErrors=[matchers.Anything()],
                errorInfo={"retryLocation": (4.0, 5.0, 6.0)},
            ),
            state_update=update_types.StateUpdate(
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id="funky-labware",
                    well_names=["A3", "A4"],
                    volume_added=update_types.CLEAR,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id", ready_to_aspirate=False
                ),
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id",
                    new_location=LabwareWellId(
                        labware_id="funky-labware", well_name="funky-well"
                    ),
                    new_deck_point=DeckPoint(x=4, y=5, z=6),
                ),
            ),
            state_update_if_false_positive=update_types.StateUpdate(),
        )



import json
import os
from typing import Any, Dict

import pytest

from .ot2_settings import (
    legacy_dummy_settings,
    migrated_dummy_settings,
    new_dummy_settings,
)
from .ot3_settings import ot3_dummy_settings
from opentrons.config import CONFIG, defaults_ot2, defaults_ot3, robot_configs
from opentrons.config.types import CurrentDict
from opentrons.hardware_control.types import BoardRevision


def test_load_corrupt_json(machine_variant_ffs: None) -> None:
    filename = os.path.join(os.path.dirname(__file__), "bad_config.json")
    with open(filename, "w") as file:
        file.write("")  # empty config file
    new_setting = robot_configs._load_json(filename)
    c = robot_configs.build_config(new_setting)
    assert c.version in [
        defaults_ot2.ROBOT_CONFIG_VERSION,
        defaults_ot3.ROBOT_CONFIG_VERSION,
    ]
    os.remove(filename)


def test_migrate_config() -> None:
    built_config = robot_configs.build_config(legacy_dummy_settings)
    new_config = robot_configs.config_to_save(built_config)
    assert new_config == migrated_dummy_settings


@pytest.mark.parametrize("config_dict", [new_dummy_settings, ot3_dummy_settings])
def test_dictify_roundtrip(config_dict: Dict[str, Any]) -> None:
    built_config = robot_configs.build_config(config_dict)
    new_saved_config = robot_configs.config_to_save(built_config)
    assert new_saved_config == config_dict


@pytest.mark.parametrize("config_dict", [new_dummy_settings, ot3_dummy_settings])
def test_json_roundtrip(config_dict: Dict[str, Any]) -> None:
    built_config = robot_configs.build_config(config_dict)
    config_dict = robot_configs.config_to_save(built_config)
    jsonstr = robot_configs.json_to_save(config_dict)
    reloaded = json.loads(jsonstr)
    rebuilt = robot_configs.build_config(reloaded)
    assert built_config == rebuilt


def test_load_legacy_gantry_cal() -> None:
    filename = CONFIG["deck_calibration_file"]
    with open(filename, "w") as file:
        deck_cal = {"gantry_calibration": [[0, 0, 0, 0]]}
        json.dump(deck_cal, file)

    result_1 = robot_configs.get_legacy_gantry_calibration()
    assert result_1 == [[0, 0, 0, 0]]

    os.remove(filename)
    result_2 = robot_configs.get_legacy_gantry_calibration()
    assert result_2 is None


@pytest.mark.parametrize(
    "current_dict,board_rev,result",
    [
        ({"default": {"X": 1}, "2.1": {"X": 2}}, BoardRevision.OG, {"X": 2}),
        ({"default": {"X": 1}, "A": {"X": 2}}, BoardRevision.OG, {"X": 1}),
        (
            {"default": {"X": 1}, "A": {"X": 2}, "2.1": {"X": 3}},
            BoardRevision.A,
            {"X": 2},
        ),
    ],
)
def test_current_for_revision(
    current_dict: CurrentDict, board_rev: BoardRevision, result: Dict[str, Any]
) -> None:
    assert robot_configs.current_for_revision(current_dict, board_rev) == result



"""Tests for the HardwareApi class."""

from typing import Dict, List, Optional, Tuple

import pytest
from _pytest.fixtures import FixtureRequest
from decoy import Decoy

from opentrons_shared_data.pipette.pipette_definition import (
    CamActionPickUpTipConfiguration,
    PressAndCamConfigurationValues,
    PressFitPickUpTipConfiguration,
)
from opentrons_shared_data.pipette.types import PipetteChannelType

from opentrons import types
from opentrons.hardware_control.instruments.ot2.pipette import Pipette
from opentrons.hardware_control.instruments.ot2.pipette_handler import (
    PipetteHandlerProvider,
)
from opentrons.hardware_control.instruments.ot3.pipette import Pipette as OT3Pipette
from opentrons.hardware_control.instruments.ot3.pipette_handler import (
    OT3PipetteHandler,
    TipActionMoveSpec,
)
from opentrons.hardware_control.types import Axis, OT3Mount


@pytest.fixture
def mock_pipette(decoy: Decoy) -> Pipette:
    return decoy.mock(cls=Pipette)


@pytest.fixture
def mock_pipette_ot3(decoy: Decoy) -> OT3Pipette:
    return decoy.mock(cls=OT3Pipette)


@pytest.fixture
def mock_pipettes_ot3(decoy: Decoy) -> Tuple[OT3Pipette, OT3Pipette]:
    return (decoy.mock(cls=OT3Pipette), decoy.mock(cls=OT3Pipette))


@pytest.fixture
def subject(decoy: Decoy, mock_pipette: Pipette) -> PipetteHandlerProvider[types.Mount]:
    inst_by_mount = {types.Mount.LEFT: mock_pipette, types.Mount.RIGHT: None}
    subject = PipetteHandlerProvider(attached_instruments=inst_by_mount)
    return subject


@pytest.fixture
def subject_ot3(decoy: Decoy, mock_pipette_ot3: OT3Pipette) -> OT3PipetteHandler:
    inst_by_mount = {OT3Mount.LEFT: mock_pipette_ot3, OT3Mount.RIGHT: None}
    subject = OT3PipetteHandler(attached_instruments=inst_by_mount)
    return subject


@pytest.fixture
def mock_presses_list() -> List[TipActionMoveSpec]:
    return [
        TipActionMoveSpec(
            distance=-10.0,
            speed=5.5,
            currents={Axis.Z_L: 1.0},
        ),
        TipActionMoveSpec(
            distance=10.0,
            speed=5.5,
            currents=None,
        ),
        TipActionMoveSpec(
            distance=-11.0,
            speed=5.5,
            currents={Axis.Z_L: 1.0},
        ),
        TipActionMoveSpec(
            distance=11.0,
            speed=5.5,
            currents=None,
        ),
    ]


@pytest.fixture
def mock_pickup_list() -> List[TipActionMoveSpec]:
    return [
        TipActionMoveSpec(
            distance=19.0,
            speed=10,
            currents={Axis.P_L: 1.0, Axis.Q: 1.0},
        ),
        TipActionMoveSpec(
            distance=29,
            speed=5.5,
            currents={Axis.P_L: 1.0, Axis.Q: 1.0},
        ),
    ]


@pytest.mark.parametrize(
    "presses_input, expected_array_length", [(0, 0), (None, 3), (3, 3)]
)
def test_plan_check_pick_up_tip_with_presses_argument(
    decoy: Decoy,
    subject: PipetteHandlerProvider[types.Mount],
    mock_pipette: Pipette,
    presses_input: int,
    expected_array_length: int,
) -> None:
    """Should return an array with expected length."""
    tip_length = 25.0
    mount = types.Mount.LEFT
    presses = presses_input
    increment = None

    decoy.when(mock_pipette.has_tip).then_return(False)
    decoy.when(mock_pipette.config.quirks).then_return([])
    decoy.when(mock_pipette.pick_up_configurations.press_fit.presses).then_return(
        expected_array_length
    )
    decoy.when(
        mock_pipette.get_pick_up_distance_by_configuration(
            mock_pipette.pick_up_configurations.press_fit
        )
    ).then_return(5)
    decoy.when(mock_pipette.pick_up_configurations.press_fit.increment).then_return(0)
    decoy.when(
        mock_pipette.get_pick_up_speed_by_configuration(
            mock_pipette.pick_up_configurations.press_fit
        )
    ).then_return(10)
    decoy.when(mock_pipette.config.end_tip_action_retract_distance_mm).then_return(0)
    decoy.when(
        mock_pipette.get_pick_up_current_by_configuration(
            mock_pipette.pick_up_configurations.press_fit
        )
    ).then_return(1.0)
    decoy.when(mock_pipette.nozzle_manager.current_configuration.tip_count).then_return(
        1
    )

    spec, _add_tip_to_instrs = subject.plan_check_pick_up_tip(
        mount=mount, tip_length=tip_length, presses=presses, increment=increment
    )

    assert len(spec.presses) == expected_array_length


@pytest.mark.parametrize(
    "presses_input, expected_array_length, channels, expected_pick_up_motor_actions",
    [
        (0, 2, 96, "mock_pickup_list"),
        (None, 4, 8, "mock_presses_list"),
        (2, 4, 1, "mock_presses_list"),
    ],
)
def test_plan_check_pick_up_tip_with_presses_argument_ot3(
    decoy: Decoy,
    subject_ot3: PipetteHandlerProvider[types.Mount],
    mock_pipette_ot3: OT3Pipette,
    presses_input: int,
    expected_array_length: int,
    channels: int,
    expected_pick_up_motor_actions: Optional[List[TipActionMoveSpec]],
    request: FixtureRequest,
) -> None:
    """Should return an array with expected length."""
    mount = OT3Mount.LEFT
    presses = presses_input
    increment = 1
    pac_values = PressAndCamConfigurationValues(
        speed=5.5, distance=10, current=1.0, tipOverlaps={"v0": {"default": 1.0}}
    )

    decoy.when(mock_pipette_ot3.has_tip).then_return(False)
    decoy.when(mock_pipette_ot3.get_pick_up_configuration()).then_return(
        CamActionPickUpTipConfiguration(
            prep_move_distance=19.0,
            prep_move_speed=10,
            configurationsByNozzleMap={"Full": {"default": pac_values}},
            connectTiprackDistanceMM=8,
        )
        if channels == 96
        else PressFitPickUpTipConfiguration(
            presses=2,
            increment=increment,
            configurationsByNozzleMap={"Full": {"default": pac_values}},
        )
    )
    decoy.when(
        mock_pipette_ot3.get_pick_up_distance_by_configuration(
            mock_pipette_ot3.get_pick_up_configuration()
        )
    ).then_return(10)
    decoy.when(
        mock_pipette_ot3.get_pick_up_speed_by_configuration(
            mock_pipette_ot3.get_pick_up_configuration()
        )
    ).then_return(5.5)
    decoy.when(
        mock_pipette_ot3.get_pick_up_current_by_configuration(
            mock_pipette_ot3.get_pick_up_configuration()
        )
    ).then_return(1.0)
    decoy.when(mock_pipette_ot3.plunger_motor_current.run).then_return(1)
    decoy.when(mock_pipette_ot3.config.quirks).then_return([])
    decoy.when(mock_pipette_ot3.channels).then_return(PipetteChannelType(channels))
    decoy.when(mock_pipette_ot3.config.end_tip_action_retract_distance_mm).then_return(
        2
    )

    assert isinstance(subject_ot3, OT3PipetteHandler)

    if channels == 96:
        spec = subject_ot3.plan_ht_pick_up_tip(96)
    else:
        spec = subject_ot3.plan_lt_pick_up_tip(mount, channels, presses, increment)
    assert len(spec.tip_action_moves) == expected_array_length
    assert spec.tip_action_moves == request.getfixturevalue(
        expected_pick_up_motor_actions
    )


def test_get_pipette_fails(
    decoy: Decoy, subject: PipetteHandlerProvider[types.Mount]
) -> None:
    with pytest.raises(types.PipetteNotAttachedError):
        subject.get_pipette(types.Mount.RIGHT)


@pytest.mark.parametrize(
    "left_offset,right_offset,ok",
    [
        (types.Point(100, 200, 300), None, True),
        (None, types.Point(-100, 200, -500), True),
        (types.Point(100, 200, 300), types.Point(200, 400, 500), False),
    ],
)
def test_ot3_pipette_handler_gives_checks_with_different_pipettes(
    left_offset: Optional[types.Point],
    right_offset: Optional[types.Point],
    ok: bool,
    mock_pipettes_ot3: Tuple[OT3Pipette, OT3Pipette],
    decoy: Decoy,
) -> None:
    """Should give you reasonable results with one or two pipettes attached."""
    # with a left and not right pipette, we should be able to pass our checks
    inst_by_mount: Dict[OT3Mount, OT3Pipette | None] = {}
    if left_offset is not None:
        inst_by_mount[OT3Mount.LEFT] = mock_pipettes_ot3[0]
        decoy.when(mock_pipettes_ot3[0].pipette_offset.offset).then_return(left_offset)
    if right_offset is not None:
        inst_by_mount[OT3Mount.RIGHT] = mock_pipettes_ot3[1]
        decoy.when(mock_pipettes_ot3[1].pipette_offset.offset).then_return(right_offset)
    subject = OT3PipetteHandler(attached_instruments=inst_by_mount)
    if left_offset is not None:
        left_result = subject.get_instrument_offset(OT3Mount.LEFT)
        assert left_result is not None
        assert left_result.offset == left_offset
        if ok:
            assert left_result.reasonability_check_failures == []
        else:
            assert len(left_result.reasonability_check_failures) == 1
    if right_offset is not None:
        right_result = subject.get_instrument_offset(OT3Mount.RIGHT)
        assert right_result is not None
        assert right_result.offset == right_offset
        if ok:
            assert right_result.reasonability_check_failures == []
        else:
            assert len(right_result.reasonability_check_failures) == 1



"""Command models to set a shake speed for a Heater-Shaker Module."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from pydantic import BaseModel, Field
from typing_extensions import Literal, Type

from ...errors.error_occurrence import ErrorOccurrence
from ...state import update_types
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from .common import get_heatershaker_ready_to_shake

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import (
        EquipmentHandler,
        MovementHandler,
        TaskHandler,
    )
    from opentrons.protocol_engine.state.state import StateView

SetShakeSpeedCommandType = Literal["heaterShaker/setShakeSpeed"]


class SetShakeSpeedParams(BaseModel):
    """Input parameters to set a shake speed for a Heater-Shaker Module."""

    moduleId: str = Field(..., description="Unique ID of the Heater-Shaker Module.")
    rpm: float = Field(..., description="Target speed in rotations per minute.")
    taskId: str | None = Field(
        None,
        description="Id for the background task that manages the temperature",
    )


class SetShakeSpeedResult(BaseModel):
    """Result data from setting and waiting for a Heater-Shaker's shake speed."""

    pipetteRetracted: bool = Field(
        ...,
        description=(
            "Whether this command automatically retracted the pipettes"
            " before starting the shake, to avoid a potential collision."
        ),
    )
    taskId: str = Field(
        description="The task id for the setTargetTemperature task",
    )


class SetShakeSpeedImpl(
    AbstractCommandImpl[SetShakeSpeedParams, SuccessData[SetShakeSpeedResult]]
):
    """Execution implementation of Heater-Shaker's set shake speed command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        task_handler: TaskHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._movement = movement
        self._task_handler = task_handler

    async def execute(
        self,
        params: SetShakeSpeedParams,
    ) -> SuccessData[SetShakeSpeedResult]:
        """Set a Heater-Shaker's target shake speed."""
        state_update = update_types.StateUpdate()

        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        hs_module_substate = self._state_view.modules.get_heater_shaker_module_substate(
            module_id=params.moduleId
        )

        validated_speed = await get_heatershaker_ready_to_shake(
            hs_module_substate, params.rpm
        )
        pipette_should_retract = (
            self._state_view.motion.check_pipette_blocking_hs_shaker(
                hs_module_substate.module_id
            )
        )
        if pipette_should_retract:
            # Move pipette away if it is close to the heater-shaker
            # TODO(jbl 2022-07-28) replace home movement with a retract movement
            await self._movement.home(
                axes=self._state_view.motion.get_robot_mount_axes()
            )
            state_update.clear_all_pipette_locations()

        # Allow propagation of ModuleNotAttachedError.
        hs_hardware_module = self._equipment.get_module_hardware_api(
            hs_module_substate.module_id
        )

        async def start_shake(task_handler: TaskHandler) -> None:
            if hs_hardware_module is not None:
                async with task_handler.synchronize_cancel_previous(
                    hs_module_substate.module_id + "-shake"
                ):
                    await hs_hardware_module.set_speed(rpm=validated_speed)

        task = await self._task_handler.create_task(
            task_function=start_shake, id=params.taskId
        )
        return SuccessData(
            public=SetShakeSpeedResult(
                pipetteRetracted=pipette_should_retract, taskId=task.id
            ),
            state_update=state_update,
        )


class SetShakeSpeed(
    BaseCommand[SetShakeSpeedParams, SetShakeSpeedResult, ErrorOccurrence]
):
    """A command to set a Heater-Shaker's shake speed."""

    commandType: SetShakeSpeedCommandType = "heaterShaker/setShakeSpeed"
    params: SetShakeSpeedParams
    result: Optional[SetShakeSpeedResult] = None

    _ImplementationCls: Type[SetShakeSpeedImpl] = SetShakeSpeedImpl


class SetShakeSpeedCreate(BaseCommandCreate[SetShakeSpeedParams]):
    """A request to create a Heater-Shaker's set shake speed command."""

    commandType: SetShakeSpeedCommandType = "heaterShaker/setShakeSpeed"
    params: SetShakeSpeedParams

    _CommandCls: Type[SetShakeSpeed] = SetShakeSpeed



from typing import Optional

import pytest

from opentrons.drivers.command_builder import CommandBuilder


def test_builder_create_command_with_terminator() -> None:
    """It should create a command with terminator."""
    terminator = "terminator"
    builder = CommandBuilder(terminator=terminator)
    assert builder.build() == "terminator"


@pytest.mark.parametrize(
    argnames=["value", "precision", "expected_float"],
    argvalues=[
        [1.2342, 3, 1.234],
        [1.2342, None, 1.2342],
        [1.2342, 0, 1.0],
    ],
)
def test_builder_create_command_add_float(
    value: float, precision: Optional[int], expected_float: float
) -> None:
    """It should create a command with a floating point value."""
    terminator = "terminator"
    builder = CommandBuilder(terminator=terminator)
    assert (
        builder.add_float(prefix="Z", value=value, precision=precision).build()
        == f"Z{expected_float} terminator"
    )


def test_builder_create_command_add_int() -> None:
    """It should create a command with an integer point value."""
    terminator = "terminator"
    builder = CommandBuilder(terminator=terminator)
    assert builder.add_int(prefix="Z", value=15).build() == f"Z15 {terminator}"


def test_builder_create_command_add_gcode() -> None:
    """It should create a command with a GCODE."""
    terminator = "terminator"
    builder = CommandBuilder(terminator=terminator)
    assert builder.add_gcode(gcode="G321").build() == f"G321 {terminator}"


def test_builder_create_command_add_builder() -> None:
    """It should create a command words in another builder."""
    terminator = "terminator"
    builder = CommandBuilder(terminator=terminator)
    assert builder.add_gcode(gcode="G321").build() == f"G321 {terminator}"

    builder2 = CommandBuilder(terminator=terminator)
    assert (
        builder2.add_builder(builder=builder)
        .add_gcode(gcode="G123")
        .add_builder(builder=builder)
        .build()
        == f"G321 G123 G321 {terminator}"
    )


def test_builder_chain() -> None:
    """It should create a command using chaining."""
    terminator = "terminator"
    builder = CommandBuilder(terminator=terminator)
    assert (
        builder.add_gcode(gcode="G321")
        .add_float(prefix="X", value=321, precision=3)
        .add_gcode(gcode="M321")
        .add_int(prefix="Z", value=3)
        .add_gcode("G111")
        .build()
        == f"G321 X321 M321 Z3 G111 {terminator}"
    )



"""Test absorbance reader initilize command."""

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition2,
    Parameters2,
)

from opentrons.drivers.types import (
    AbsorbanceReaderLidStatus,
)
from opentrons.hardware_control.modules import AbsorbanceReader
from opentrons.protocol_engine import DeckSlotLocation, ModuleModel
from opentrons.protocol_engine.commands.absorbance_reader import (
    OpenLidParams,
    OpenLidResult,
)
from opentrons.protocol_engine.commands.absorbance_reader.open_lid import (
    OpenLidImpl,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.errors import CannotPerformModuleAction
from opentrons.protocol_engine.execution import EquipmentHandler, LabwareMovementHandler
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.module_substates import (
    AbsorbanceReaderId,
    AbsorbanceReaderSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.types import DeckSlotName


@pytest.fixture
def absorbance_def() -> LabwareDefinition2:
    """Get a tip rack Pydantic model definition value object."""
    return LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        namespace="test",
        version=1,
        parameters=Parameters2.model_construct(  # type: ignore[call-arg]
            loadName="cool-labware",
            tipOverlap=None,  # add a None value to validate serialization to dictionary
        ),
    )


@pytest.fixture
def subject(
    state_view: StateView,
    equipment: EquipmentHandler,
    labware_movement: LabwareMovementHandler,
) -> OpenLidImpl:
    """Command implementation subject for testing."""
    return OpenLidImpl(
        state_view=state_view, equipment=equipment, labware_movement=labware_movement
    )


@pytest.mark.parametrize(
    "hardware_lid_status",
    (AbsorbanceReaderLidStatus.ON, AbsorbanceReaderLidStatus.OFF),
)
async def test_absorbance_reader_implementation(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    labware_movement: LabwareMovementHandler,
    hardware_lid_status: AbsorbanceReaderLidStatus,
    absorbance_def: LabwareDefinition2,
    subject: OpenLidImpl,
) -> None:
    """It should validate, find hardware module if not virtualized, and disengage."""
    params = OpenLidParams(
        moduleId="unverified-module-id",
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)
    verified_module_id = AbsorbanceReaderId("module-id")

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when(await absorbance_module_hw.get_current_lid_status()).then_return(
        hardware_lid_status
    )
    decoy.when(state_view.modules.get_requested_model(params.moduleId)).then_return(
        ModuleModel.ABSORBANCE_READER_V1
    )

    decoy.when(state_view.labware.get_absorbance_reader_lid_definition()).then_return()

    decoy.when(state_view.modules.get_location(params.moduleId)).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_D3)
    )
    decoy.when(
        state_view.modules.ensure_and_convert_module_fixture_location(
            DeckSlotName.SLOT_D3, ModuleModel.ABSORBANCE_READER_V1
        )
    ).then_return("absorbanceReaderV1D3")

    decoy.when(state_view.labware.get_absorbance_reader_lid_definition()).then_return(
        absorbance_def
    )

    result = await subject.execute(params=params)

    assert result == SuccessData(
        public=OpenLidResult(),
        state_update=update_types.StateUpdate(
            absorbance_reader_state_update=update_types.AbsorbanceReaderStateUpdate(
                module_id="module-id",
                absorbance_reader_lid=update_types.AbsorbanceReaderLidUpdate(
                    is_lid_on=False
                ),
            ),
        ),
    )


async def test_open_lid_raises_no_module(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: OpenLidImpl,
) -> None:
    """Should raise an error that the hardware module not found."""
    params = OpenLidParams(
        moduleId="unverified-module-id",
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    verified_module_id = AbsorbanceReaderId("module-id")

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)
    decoy.when(state_view.config.use_virtual_modules).then_return(False)
    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(None)
    with pytest.raises(CannotPerformModuleAction):
        await subject.execute(params=params)



import pytest
from mock import AsyncMock

from opentrons.drivers.smoothie_drivers import SmoothieDriver
from opentrons.tools import write_pipette_memory


@pytest.fixture
def mock_driver() -> AsyncMock:
    return AsyncMock(spec=SmoothieDriver)


async def test_write_identifiers(mock_driver: AsyncMock) -> None:
    """It should call driver to write a new id and model."""
    mount = "left"
    new_id = "some id"
    new_model = "some model"

    mock_driver.read_pipette_id.return_value = new_id
    mock_driver.read_pipette_model.return_value = new_model

    await write_pipette_memory.write_identifiers(
        mount=mount, new_id=new_id, new_model=new_model, driver=mock_driver
    )

    mock_driver.write_pipette_id.assert_called_once_with(mount, new_id)
    mock_driver.read_pipette_id.assert_called_once_with(mount)
    mock_driver.write_pipette_model.assert_called_once_with(mount, new_model)
    mock_driver.read_pipette_model.assert_called_once_with(mount)


async def test_write_identifiers_id_mismatch(mock_driver: AsyncMock) -> None:
    """It should fail when written id doesn't match read id."""
    mount = "left"
    new_id = "some id"
    new_model = "some model"

    mock_driver.read_pipette_id.return_value = new_id + "_wrong"

    with pytest.raises(Exception):
        await write_pipette_memory.write_identifiers(
            mount=mount, new_id=new_id, new_model=new_model, driver=mock_driver
        )


async def test_write_identifiers_model_mismatch(mock_driver: AsyncMock) -> None:
    """It should fail when written model doesn't match read model."""
    mount = "left"
    new_id = "some id"
    new_model = "some model"

    mock_driver.read_pipette_id.return_value = new_id
    mock_driver.read_pipette_model.return_value = new_model + "_wrong"

    with pytest.raises(Exception):
        await write_pipette_memory.write_identifiers(
            mount=mount, new_id=new_id, new_model=new_model, driver=mock_driver
        )


async def test_check_previous_data(mock_driver: AsyncMock) -> None:
    """It should read the pipette id and model"""
    mount = "left"

    await write_pipette_memory.check_previous_data(mount, mock_driver)

    mock_driver.read_pipette_id.assert_called_once_with(mount)
    mock_driver.read_pipette_model.assert_called_once_with(mount)


pipette_barcode_to_model = {
    "P10S20180101A01": "p10_single_v1",
    "P10M20180101A01": "p10_multi_v1",
    "P50S180101A01": "p50_single_v1",
    "P50M20180101B01": "p50_multi_v1",
    "P300S20180101A01": "p300_single_v1",
    "P300M20180101A01": "p300_multi_v1",
    "P1000S20180101A01": "p1000_single_v1",
    "P10SV1318010101": "p10_single_v1.3",
    "P10MV1318010102": "p10_multi_v1.3",
    "P50SV1318010103": "p50_single_v1.3",
    "P50MV1318010104": "p50_multi_v1.3",
    "P3HSV1318010105": "p300_single_v1.3",
    "P3HMV1318010106": "p300_multi_v1.3",
    "P1KSV1318010107": "p1000_single_v1.3",
    "P10SV1418010101": "p10_single_v1.4",
    "P10MV1418010102": "p10_multi_v1.4",
    "P50SV1418010103": "p50_single_v1.4",
    "P50MV1418010104": "p50_multi_v1.4",
    "P3HSV1418010105": "p300_single_v1.4",
    "P3HMV1418010106": "p300_multi_v1.4",
    "P1KSV1418010107": "p1000_single_v1.4",
    "P20MV2120120204": "p20_multi_v2.1",
    "P1KSV2218010107": "p1000_single_v2.2",
    "P20SV2220020501": "p20_single_v2.2",
}


def test_parse_model_from_barcode() -> None:
    for barcode, model in pipette_barcode_to_model.items():
        assert write_pipette_memory._parse_model_from_barcode(barcode) == model

    with pytest.raises(Exception):
        write_pipette_memory._parse_model_from_barcode("P1HSV1318010101")

    with pytest.raises(Exception):
        write_pipette_memory._parse_model_from_barcode("P1KSV1218010101")

    with pytest.raises(Exception):
        write_pipette_memory._parse_model_from_barcode("aP300S20180101A01")
